package gov.usgs.earthquake.nshmp.gmm;

import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.MW;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.RRUP;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.VS30;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.ZHYP;
import static gov.usgs.earthquake.nshmp.gmm.Imt.PGA;
import static java.lang.Math.exp;
import static java.lang.Math.log;
import static java.lang.Math.log10;
import static java.lang.Math.max;
import static java.lang.Math.min;
import static java.lang.Math.pow;
import static java.lang.Math.sqrt;

import java.util.Map;

import com.google.common.collect.Range;

import gov.usgs.earthquake.nshmp.Maths;
import gov.usgs.earthquake.nshmp.gmm.GmmInput.Constraints;

/**
 * Implementation of the PEER NGA-Subduction GMM, Parker et al., 2020 GMM
 * (PSHAB20)
 *
 * Notes:
 *
 * This model is only applicable to sites in the forearc, future residuals
 * analysis is planned to determine if there are attenuation differences between
 * forearc and backarc zones.
 *
 * PSHAB20 uses zHyp (randomized source location on fault), but this can be
 * replaced this event-specific zHyp with the mean depth expected for the fault
 * plane, zHypBar, in Eq. 4.6 without modification to coefficients or
 * uncertainty.
 *
 * Path term (Eqs. 4.2 and 4.3 in PEER report) appears to differ in the official
 * Python implementation. Compare implementations of getPathTerm() below
 *
 * <p><b>Reference:</b> Parker, G.A, Stewart, J.P., Boore, D.M., Atkinson, G.M.,
 * Hassani, B. (2020) NGA-Subduction global ground-motion models with regional
 * adjustment factors. Report no. 2020/03. Berkeley, CA: PEER, 131 p.
 *
 * <p><b>Component:</b> TODO
 *
 * @author U.S. Geological Survey
 */

public abstract class ParkerEtAl_2020 implements GroundMotionModel {

  static final String NAME = "Parker et al. (2020) NGA-Subduction";

  static final Constraints CONSTRAINTS_INTERFACE = Constraints.builder()
      .set(MW, Range.closed(4.5, 9.5))
      .set(RRUP, Range.closed(20.0, 1000.0))
      .set(ZHYP, Range.closed(0.0, 40.0))
      .set(VS30, Range.closed(150.0, 2000.0))
      .build();
  static final Constraints CONSTRAINTS_SLAB = Constraints.builder()
      .set(MW, Range.closed(4.5, 8.5))
      .set(RRUP, Range.closed(35.0, 1000.0))
      .set(ZHYP, Range.closed(0.0, 200.0))
      .set(VS30, Range.closed(150.0, 2000.0))
      .build();

  /*
   * Regions: global, Alaska, Cascadia, [CAM (Central America), Japan, SA (South
   * America), Taiwan]
   *
   * Saturation Regions: global, Aleutian, Alaska, Cascadia, [Central America S,
   * Central America N, Japan Pacific, Japan Phillipines, South America N, South
   * America S, Taiwan W, Taiwan E]
   */
  static final class Region {
    static final String GLOBAL = "Global";
    static final String ALASKA = "Alaska";
    static final String ALEUTIANS = "Aleutians";
    static final String CASCADIA = "Cascadia";
  }

  static final class Basin {
    static final String NONE = "None";
    static final String PACNW_NONE = "PacNW-no-basin";
    static final String PACNW = "PacNW";
    static final String SEATTLE = "Seattle";
  }

  // aleatory uncertainty coefficients are duplicated in interface and interface
  // CSV files (as are some other coefficients that are also independent of
  // event type)
  private static final String INTERFACE_CSV = "PSHAB20_interface.csv";
  private static final String SLAB_CSV = "PSHAB20_slab.csv";
  // 4 epistemic coefficients tied to region and interface/slab
  private static final String EPISTEMIC_INTERFACE_CSV = "PSHAB20_epistemic_interface.csv";
  private static final String EPISTEMIC_SLAB_CSV = "PSHAB20_epistemic_slab.csv";

  static final CoefficientContainer COEFFS_INTERFACE;
  static final CoefficientContainer COEFFS_SLAB;

  static {
    COEFFS_INTERFACE = new CoefficientContainer(INTERFACE_CSV);
    COEFFS_SLAB = new CoefficientContainer(SLAB_CSV);
  }

  // private static final double B4 = 0.1;
  // private static final double V1 = 270.0; // m/s
  // private static final double VREF = 760.0; // m/s
  private static final double DB1 = 20.0; // km
  private static final double DB2 = 67.0; // km

  /* ϕ parameters */
  private static final double R1 = 200; // km
  private static final double R2 = 500; // km
  private static final double V1 = 200; // m/s
  private static final double V2 = 500; // m/s
  /* ϕS2S and ϕSS parameters */
  private static final double R3 = 200; // km
  private static final double R4 = 500; // km
  private static final double R5 = 500; // km
  private static final double R6 = 800; // km
  private static final double V3 = 200; // m/s
  private static final double V4 = 800; // m/s

  private static final class Coefficients {
    final Imt imt;
    final double c0; // model constant
    final double c1, b4, a0; // path scaling
    final double mc; // saturation corner magnitude
    final double c4, c5, c6; // magnitude scaling
    final double d, m, db; // source depth scaling (slab only)
    final double v1, v2, vRef, s1, s2; // linear site amplification
    final double f4, f5; // non-linear site amplification
    final double c_e1, c_e2, c_e3, del_None, del_Seattle; // Cascadia basin
    final double τ, ϕ21, ϕ22, ϕ2v, vm, ϕ2s2s0, a1, ϕ2ss1, ϕ2ss2, a2; // alleatory
                                                                     // uncertainty
    final double σε1, σε2, t1, t2;

    Coefficients(Imt imt, CoefficientContainer cc, boolean isSlab, String region) {
      this.imt = imt;
      Map<String, Double> coeffs = cc.get(imt);
      // Model constant
      c0 = coeffs.get(region + "_c0");
      // Path
      c1 = coeffs.get("c1");
      b4 = coeffs.get("b4"); // period-independent
      a0 = coeffs.get(getA0Col(isSlab, region));
      // a0 = region.equals(Region.ALASKA) || (region.equals(Region.CASCADIA) &&
      // isSlab)
      // ? coeffs.get(region + "_a0") : coeffs.get(Region.GLOBAL + "_a0");
      // Magnitude saturation corner magnitude
      mc = getSaturationMagnitude(region, isSlab);
      // Magnitude-scaling
      c4 = coeffs.get("c4");
      c5 = coeffs.get("c5");
      c6 = coeffs.get("c6");
      // Source depth scaling (slab only)
      d = isSlab ? coeffs.get("d") : 0.0;
      m = isSlab ? coeffs.get("m") : 0.0;
      db = isSlab ? coeffs.get("db") : 0.0;
      // Linear site amplification
      v1 = coeffs.get("V1");
      v2 = coeffs.get("V2");;
      vRef = coeffs.get("Vref");
      s2 = (region.equals(Region.ALEUTIANS)) ? coeffs.get(Region.GLOBAL + "_s2")
          : coeffs.get(region + "_s2");
      s1 = s2; // s1 = s2 except for Japan and Taiwan
      // Non-linear site amplification
      f4 = coeffs.get("f4");
      f5 = coeffs.get("f5");
      // Cascadia basin
      c_e1 = coeffs.get("C_e1");
      c_e2 = coeffs.get("C_e2");
      c_e3 = coeffs.get("C_e3");
      del_None = coeffs.get("del_None");
      del_Seattle = coeffs.get("del_Seattle");
      // aleatory model
      τ = coeffs.get("Tau");
      ϕ21 = coeffs.get("phi21");
      ϕ22 = coeffs.get("phi22");
      ϕ2v = coeffs.get("phi2V");
      vm = coeffs.get("VM");
      ϕ2s2s0 = coeffs.get("phi2S2S0");
      a1 = coeffs.get("a1");
      ϕ2ss1 = coeffs.get("phi2SS1");
      ϕ2ss2 = coeffs.get("phi2SS2");
      a2 = coeffs.get("a2");
      // epistemic model (period-independent)
      σε1 = getσε1(region, isSlab);
      σε2 = getσε2(region, isSlab);
      t1 = getT1(region, isSlab);
      t2 = getT2(region, isSlab);
    }

    private static final String getA0Col(boolean isSlab, String region) {
      String a0Col;
      if (region.equals(Region.ALASKA)) {
        a0Col = Region.ALASKA + "_a0";
      } else if (region.equals(Region.CASCADIA) && isSlab) {
        a0Col = Region.CASCADIA + "_a0";
      } else {
        a0Col = Region.GLOBAL + "_a0";
      }
      return a0Col;
    }

    /* Table E4 */
    private static final double getσε1(String region, boolean isSlab) {
      // return region.equals(Region.ALASKA) ? 0.15 :
      // region.equals(Region.ALEUTIANS) ? 0.15 :
      // region.equals(Region.CASCADIA) ? 0.43 : 0.4; // else GLOBAL
      switch (region) {
        case Region.GLOBAL:
          return isSlab ? 0.35 : 0.4;
        case Region.ALASKA:
          return isSlab ? 0.15 : .15;
        case Region.ALEUTIANS:
          return isSlab ? 0.25 : .15;
        case Region.CASCADIA:
          return isSlab ? 0.35 : 0.43;
        default:
          throw new IllegalArgumentException("Region [" + region + "] not supported");
      }
    }

    /* Table E4 */
    private static final double getσε2(String region, boolean isSlab) {
      // return region.equals(Region.ALASKA) ? 0.1 :
      // region.equals(Region.ALEUTIANS) ? 0.11 :
      // region.equals(Region.CASCADIA) ? 0.33 : 0.4; // else GLOBAL
      switch (region) {
        case Region.GLOBAL:
          return isSlab ? 0.22 : 0.4;
        case Region.ALASKA:
          return isSlab ? 0.12 : .1;
        case Region.ALEUTIANS:
          return isSlab ? 0.18 : .1;
        case Region.CASCADIA:
          return isSlab ? 0.16 : 0.33;
        default:
          throw new IllegalArgumentException("Region [" + region + "] not supported");
      }
    }

    /* Table E4 */
    private static final double getT1(String region, boolean isSlab) {
      // return region.equals(Region.ALASKA) ? 1 :
      // region.equals(Region.ALEUTIANS) ? 1 :
      // region.equals(Region.CASCADIA) ? 0.2 : 0.2; // else GLOBAL
      switch (region) {
        case Region.GLOBAL:
          return isSlab ? 0.15 : 0.2;
        case Region.ALASKA:
          return isSlab ? 0.5 : 1;
        case Region.ALEUTIANS:
          return isSlab ? 0.3 : 1;
        case Region.CASCADIA:
          return isSlab ? 0.2 : 0.2;
        default:
          throw new IllegalArgumentException("Region [" + region + "] not supported");
      }
    }

    /* Table E4 */
    private static final double getT2(String region, boolean isSlab) {
      switch (region) {
        case Region.GLOBAL:
          return isSlab ? 2 : 0.4;
        case Region.ALASKA:
          return isSlab ? 1 : 4;
        case Region.ALEUTIANS:
          return isSlab ? 0.8 : 4;
        case Region.CASCADIA:
          return isSlab ? 3 : 0.5;
        default:
          throw new IllegalArgumentException("Region [" + region + "] not supported");
      }
    }

    private static final double getSaturationMagnitude(String region, boolean isSlab) {
      // TODO? Read from coeff table?
      // Parket et al. (2020) Table 4.1
      // Regional saturation magnitudes for interface events computed using
      // seismogenic fault width [Campbell 2020] and for intraslab events
      // computed
      // using slab thickness [Ji and Archuleta 2018].
      switch (region) {
        case Region.GLOBAL:
          return isSlab ? 7.6 : 7.9;
        case Region.ALASKA:
          return isSlab ? 7.20 : 8.6;
        case Region.ALEUTIANS:
          return isSlab ? 7.98 : 8.0;
        case Region.CASCADIA:
          return isSlab ? 7.20 : 7.7;
        default:
          throw new IllegalArgumentException("Region [" + region + "] not supported");
      }
    }
  }

  private final Coefficients coeffs;
  private final Coefficients coeffsPGA;

  ParkerEtAl_2020(final Imt imt) {
    coeffs = initCoeffs(imt, isSlab(), getRegion());
    coeffsPGA = initCoeffs(PGA, isSlab(), getRegion());
  }

  private static Coefficients initCoeffs(final Imt imt, final boolean isSlab, final String region) {
    CoefficientContainer coeffs = isSlab ? COEFFS_SLAB : COEFFS_INTERFACE;
    return new Coefficients(imt, coeffs, isSlab, region);
  }

  @Override
  public final ScalarGroundMotion calc(final GmmInput in) {
    /*
     * Equation 4.1 μ_lny = c0 + Fp + Fm + Fd + Fs
     */

    // double Mc = getSaturationMagnitude(getRegion(), isSlab());
    double pgaRef = exp(calcMean(coeffsPGA, isSlab(), in.Mw, in.rRup, in.zHyp));

    double μ = calcMean(coeffs, isSlab(), getBasin(), in.Mw, in.rRup, in.zHyp, pgaRef,
        in.vs30, in.z2p5);

    double ϕTotal = getϕTotal(coeffs, in.rRup, in.vs30);
    double σ = Maths.hypot(coeffs.τ, ϕTotal);

    // System.out.println("calc for T = " + coeffs.imt.toString());
    // System.out.printf(" ϕTotal = %.6f\n", ϕTotal);
    // System.out.printf(" σ = %.6f\n", σ);

    double ϕS2S = getϕS2S(coeffs, in.rRup, in.vs30);
    double ϕSS = getϕSS(coeffs, in.rRup, in.vs30);

    double σε = getσε(coeffs);

    return DefaultScalarGroundMotion.create(μ, σ);
  }

  abstract boolean isSlab();

  abstract String getRegion();

  abstract String getBasin();

  // abstract double σε1();
  //
  // abstract double σε2();
  //
  // abstract double t1();
  //
  // abstract double t2();

  /* calc median at 760 reference conditions */
  private static final double calcMean(
      final Coefficients c,
      final boolean isSlab,
      final double Mw,
      final double rRup,
      final double zHyp) {
    double h = nearSourceSaturation(Mw, c.mc, isSlab);
    double Fp = getPathTermPy(c, rRup, Mw, h);
    double Fm = getMagnitudeTerm(c, Mw);
    double Fd = getDepthTerm(c, isSlab, zHyp);

    // System.out.println("calcMean Reference for T = " + c.imt.toString());
    // System.out.printf(" mc = %.2f\n", c.mc);
    // System.out.printf(" h = %.6f\n", h);
    // System.out.printf(" Fp = %.6f\n", Fp);
    // System.out.printf(" Fm(ref) = %.6f\n", Fm);
    // System.out.printf(" Fd = %.6f\n", Fd);
    // System.out.printf(" Fb = %.6f (reference calc)\n", 0.0);
    // System.out.printf(" mu = %.6f\n", c.c0 + Fp + Fm + Fd);

    return c.c0 + Fp + Fm + Fd;
  }

  /* Equation 4.1 */
  private static final double calcMean(
      final Coefficients c,
      final boolean isSlab,
      final String basin,
      final double Mw,
      final double rRup,
      final double zHyp,
      final double pgaRef,
      final double vs30,
      final double z2p5) {

    double h = nearSourceSaturation(Mw, c.mc, isSlab);
    double Fp = getPathTermPy(c, rRup, Mw, h);
    double Fm = getMagnitudeTerm(c, Mw);
    double Fd = getDepthTerm(c, isSlab, zHyp);

    /* Equation 4.7: Fs = Flin + Fnl + Fb */
    double Fslin = getSiteTermLinear(c, vs30);
    double Fsnl = getSiteTermNonLinear(c, pgaRef, vs30);
    double Fsb = getSiteBasinTerm(c, basin, vs30, z2p5);
    double Fs = Fslin + Fsnl + Fsb;

    // System.out.println("calcMean for T = " + c.imt.toString());
    // System.out.printf(" Vs30 = %.2f\n", vs30);
    // System.out.printf(" h = %.6f\n", h);
    // System.out.printf(" Fp = %.6f\n", Fp);
    // System.out.printf(" Fm = %.6f\n", Fm);
    // System.out.printf(" Fd = %.6f\n", Fd);
    // System.out.printf(" PGAref = %.6f\n", pgaRef);
    // System.out.printf(" Fslin = %.6e\n", Fslin);
    // System.out.printf(" Fsnl = %.6e\n", Fsnl);
    // System.out.printf(" Fssb = %.6e\n", Fsb);
    // System.out.printf(" Fs = %.6e\n", Fs);
    // System.out.printf(" mu = %.6f\n", c.c0 + Fp + Fm + Fd + Fs);
    // System.out.printf(" y = %.6f\n", exp(c.c0 + Fp + Fm + Fd + Fs));

    return c.c0 + Fp + Fm + Fd + Fs;
  }

  /* Equations 4.2 and 4.3 */
  private static final double getPathTerm(final Coefficients c, double rRup, double Mw, double h) {
    double R = Maths.hypot(rRup, h);
    return c.c1 * log(R) + c.b4 * Mw * log(rRup / R) + c.a0 * R;
  }

  /* Equations 4.2 and 4.3, following official Python implementation */
  private static final double getPathTermPy(final Coefficients c, double rRup, double Mw,
      double h) {
    double Rref = Maths.hypot(1, h);
    double R = Maths.hypot(rRup, h);
    double R_Rref = R / Rref;
    // System.out.printf(" c1 = %.6e\n", c.c1);
    // System.out.printf(" b4 = %.6e\n", c.b4);
    // System.out.printf(" a0 = %.6e\n", c.a0);
    // System.out.printf(" Rref = %.6e\n", R_Rref);
    // System.out.printf(" R = %.6e\n", R);
    // System.out.printf(" Mw = %.6e\n", Mw);

    return c.c1 * log(R) + c.b4 * Mw * log(R_Rref) + c.a0 * R;
  }

  private static final double nearSourceSaturation(double Mw, double mc, boolean slab) {
    if (slab) {
      /* Equation 4.4b */
      return (Mw <= mc) ? pow(10, (1.050 / (mc - 4.0)) * (Mw - mc) + 1.544) : 35;
    } else {
      /* Equation 4.4a */
      return pow(10.0, -0.82 + 0.252 * Mw);
    }
  }

  /* Equation 4.5 */
  private static final double getMagnitudeTerm(final Coefficients c, double Mw) {
    double dM = Mw - c.mc;
    return (Mw <= c.mc) ? c.c4 * dM + c.c5 * dM * dM : c.c6 * dM;
  }

  /* Equation 4.6 */
  private static final double getDepthTerm(final Coefficients c, boolean slab, double zHyp) {
    if (!slab) {
      return 0.0;
    }
    // Zhyp (see section 4.3.3) can be replaced with mean depth (Zhyp_bar)
    // that depends on Ztor, fault width, and dip angle
    double dTerm;
    if (zHyp < DB1) { // "less than"
      dTerm = c.m * (DB1 - DB2) + c.d;
    } else if (zHyp > DB2) {
      dTerm = c.d;
    } else { // "db1 <= zHyp" possible typo? (written as "db1 < xHyp")
      dTerm = c.m * (zHyp - DB2) + c.d;
    }
    return dTerm;
  }

  /* Equation 4.8 */
  private static final double getSiteTermLinear(final Coefficients c, double vs30) {
    double Flin;
    if (vs30 <= c.v1) {
      Flin = c.s1 * log(vs30 / c.v1) + c.s2 * log(c.v1 / c.vRef);
    } else if (vs30 > c.v2) {
      Flin = c.s2 * log(c.v2 / c.vRef);
    } else {
      Flin = c.s2 * log(vs30 / c.vRef);
    }
    return Flin;
  }

  /* Equation 4.9 */
  private static final double getSiteTermNonLinear(final Coefficients c, double pgaRef,
      double vs30) {
    double f3 = 0.05;
    double f2 = c.f4 * (exp(c.f5 * (min(vs30, 760) - 200)) - exp(c.f5 * (760 - 200)));
    return 0.0 + f2 * log((pgaRef + f3) / f3);
  }

  /* Equations 4.11/5.5 - 4.13/5.7 */
  private static final double getSiteBasinTerm(final Coefficients c, String basin, double vs30,
      double z2p5) {
    if (basin.equals(Basin.NONE) || Double.isNaN(z2p5)) {
      return 0.0;
    }
    // Cascadia coefficients
    double θ0 = 3.94;
    double θ1 = -0.42;
    double νμ = 200.0;
    double νσ = 0.2;

    double e1 = c.c_e1;
    double e2;
    double e3;

    if (basin.equals(Basin.PACNW_NONE)) {
      e2 = c.c_e2 + c.del_None;
      e3 = c.c_e3 + c.del_None;
    } else if (basin.equals(Basin.SEATTLE)) {
      e2 = c.c_e2 + c.del_Seattle;
      e3 = c.c_e3 + c.del_Seattle;
    } else if (basin.equals(Basin.PACNW)) {
      e2 = c.c_e2;
      e3 = c.c_e3;
    } else {
      throw new IllegalArgumentException("Basin [" + basin + "] not supported");
    }

    // Eq. 4.13/5.7
    double μz2p5 = exp(
        log(10) * θ1 * (1 + Maths.erf((log10(vs30) - log10(νμ)) / νσ / sqrt(2)) + log(10) * θ0));
    double δz2p5 = log(z2p5) - log(μz2p5);

    double Fb;
    if (δz2p5 <= (e1 / e3)) {
      Fb = e1;
    } else if (δz2p5 >= (e2 / e3)) {
      Fb = e2;
    } else {
      Fb = e3 * δz2p5;
    }
    return Fb;
  }

  /* Eqs. 6.3, 6.4, 6.5 */
  private static final double getϕTotal(final Coefficients c, final double rRup,
      final double vs30) {

    /* Eq. 6.5 */
    double Δvar;
    double rPrime = max(R1, min(R2, rRup));
    if (vs30 <= V1) {
      Δvar = c.ϕ2v * (log(R2 / rPrime) / log(R2 / R1));
    } else if (vs30 >= V2) {
      Δvar = 0;
    } else {
      Δvar = c.ϕ2v * (log(V2 / vs30) / log(V2 / V1)) * (log(R2 / rPrime) / log(R2 / R1));
    }

    /* Eq. 6.4 */
    double ϕ2;
    if (rRup <= R1) {
      ϕ2 = c.ϕ21;
    } else if (rRup >= R2) {
      ϕ2 = c.ϕ22;
    } else {
      ϕ2 = (c.ϕ22 - c.ϕ21) / log(R2 / R1) * log(rRup / R1) + c.ϕ21;
    }

    /* Eq. 6.3 */
    double ϕTotal = sqrt(Δvar + ϕ2);

    // System.out.println("getϕTotal for T = " + c.imt.toString());
    // System.out.printf(" rRup = %.6f\n",rRup);
    // System.out.printf(" vs30 = %.6f\n",vs30);
    // System.out.printf(" rPrime = %.6f\n",rPrime);
    // System.out.printf(" c.ϕ2v = %.6f\n",c.ϕ2v);
    // System.out.printf(" c.ϕ21 = %.6f\n",c.ϕ21);
    // System.out.printf(" c.ϕ22 = %.6f\n",c.ϕ22);
    // System.out.printf(" Δvar = %.6f\n",Δvar);
    // System.out.printf(" ϕ2 = %.6f\n",ϕ2);
    // System.out.printf(" ϕTotal = %.6f\n",ϕTotal);

    return ϕTotal;
  }

  /* Eqs. 6.6, 6.7 */
  private static final double getϕS2S(final Coefficients c, final double rRup, final double vs30) {

    /* Eq. 6.7 */
    double rPrime = max(R3, min(R4, rRup));
    double ΔvarS2S;
    if (vs30 <= V3) {
      ΔvarS2S = c.a1 * log(V3 / c.vm) * log(R4 / rPrime) / log(R4 / R3);
    } else if (vs30 < c.vm) {
      ΔvarS2S = c.a1 * log(vs30 / c.vm) * log(R4 / rPrime) / log(R4 / R3);
    } else if (vs30 < V4) {
      ΔvarS2S = c.a1 * log(vs30 / c.vm);
    } else { // vs30 >= V4
      ΔvarS2S = c.a1 * log(V4 / c.vm);
    }

    double ϕS2S = sqrt(c.ϕ2s2s0 + ΔvarS2S);

    // System.out.println("getϕS2S for T = " + c.imt.toString());
    // System.out.printf(" rRup = %.6f\n",rRup);
    // System.out.printf(" vs30 = %.6f\n",vs30);
    // System.out.printf(" rPrime = %.6f\n",rPrime);
    // System.out.printf(" c.a1 = %.6f\n",c.a1);
    // System.out.printf(" c.vm = %.6f\n",c.vm);
    // System.out.printf(" ΔvarS2S = %.6f\n",ΔvarS2S);
    // System.out.printf(" c.ϕ2s2s0 = %.6f\n",c.ϕ2s2s0);
    // System.out.printf(" ϕS2S = %.6f\n",ϕS2S);

    return ϕS2S;
  }

  /* Eqs. 6.8, 6.9, 6.10 */
  private double getϕSS(Coefficients c, double rRup, double vs30) {
    /* Eq. 6.10 */
    double rPrime = max(R3, min(R4, rRup));
    double ΔvarSS;
    if (vs30 <= V3) {
      ΔvarSS = c.a2 * log(V3 / c.vm) * log(R4 / rPrime) / log(R4 / R3);
    } else if (vs30 < c.vm) {
      ΔvarSS = c.a2 * log(vs30 / c.vm) * log(R4 / rPrime) / log(R4 / R3);
    } else if (vs30 < V4) {
      ΔvarSS = c.a2 * log(vs30 / c.vm);
    } else { // vs30 >= V4
      ΔvarSS = c.a2 * log(V4 / c.vm);
    }

    /* Eq. 6.9 */
    double ϕSS2rRup;
    if (rRup <= R5) {
      ϕSS2rRup = c.ϕ2ss1;
    } else if (rRup < R6) {
      ϕSS2rRup = (c.ϕ2ss2 - c.ϕ2ss1) / log(R6 / R5) * log(rRup / R5) + c.ϕ2ss1;
    } else { // rRup >= R6
      ϕSS2rRup = c.ϕ2ss2;
    }

    double ϕSS = sqrt(ϕSS2rRup + ΔvarSS);

    // System.out.println("getϕSS for T = " + c.imt.toString());
    // System.out.printf(" rRup = %.6f\n",rRup);
    // System.out.printf(" vs30 = %.6f\n",vs30);
    // System.out.printf(" rPrime = %.6f\n",rPrime);
    // System.out.printf(" c.a2 = %.6f\n",c.a1);
    // System.out.printf(" c.vm = %.6f\n",c.vm);
    // System.out.printf(" c.ϕ2ss1 = %.6f\n",c.ϕ2ss1);
    // System.out.printf(" c.ϕ2ss2 = %.6f\n",c.ϕ2ss2);
    // System.out.printf(" ΔvarSS = %.6f\n",ΔvarSS);
    // System.out.printf(" ϕSS2rRup = %.6f\n",ϕSS2rRup);
    // System.out.printf(" ϕS2S = %.6f\n",ϕSS);

    return ϕSS;
  }

  /* Eq. 8.2 */
  private double getσε(Coefficients c) {
    double σε;
    if (c.imt == Imt.PGA || c.imt == Imt.PGV || c.imt.period() < c.t1) {
      σε = c.σε1;
    } else if (c.imt.period() < c.t2) {
      σε = c.σε1 - (c.σε1 - c.σε2) * log(c.imt.period() / c.t1) / log(c.t2 / c.t1);
    } else { // T > T2
      σε = c.σε2;
    }
    // System.out.println("getσε for T = " + c.imt.toString());
    // System.out.printf(" T1 = %.6f\n",c.t1);
    // System.out.printf(" T2 = %.6f\n",c.t2);
    // System.out.printf(" c.σε1 = %.6f\n",c.σε1);
    // System.out.printf(" c.σε2 = %.6f\n",c.σε2);
    // System.out.printf(" σε = %.6f\n",σε);

    return σε;
  }

  /*
   * Global interface model
   */
  static class Interface extends ParkerEtAl_2020 {
    static final String NAME = ParkerEtAl_2020.NAME + " : Interface";

    Interface(Imt imt) {
      super(imt);
    }

    @Override
    final boolean isSlab() {
      return false;
    }

    @Override
    String getRegion() {
      return Region.GLOBAL;
    }

    @Override
    String getBasin() {
      return Basin.NONE;
    }

    // @Override
    // double σε1() {
    // return 0;
    // }
    //
    // @Override
    // double σε2() {
    // return 0;
    // }
    //
    // @Override
    // double t1() {
    // return 0;
    // }
    //
    // @Override
    // double t2() {
    // return 0;
    // }

  }

  /*
   * Global intraslab model
   */
  static class Slab extends ParkerEtAl_2020 {
    static final String NAME = ParkerEtAl_2020.NAME + " : Slab";

    Slab(Imt imt) {
      super(imt);
    }

    @Override
    final boolean isSlab() {
      return true;
    }

    @Override
    String getRegion() {
      return Region.GLOBAL;
    }

    @Override
    String getBasin() {
      return Basin.NONE;
    }

    // @Override
    // double σε1() {
    // return 0;
    // }
    //
    // @Override
    // double σε2() {
    // return 0;
    // }
    //
    // @Override
    // double t1() {
    // return 0;
    // }
    //
    // @Override
    // double t2() {
    // return 0;
    // }
  }

  /*
   * Cascadia interface model
   */
  static class CascadiaInterface extends Interface {
    static final String NAME = Interface.NAME + " : Cascadia";

    CascadiaInterface(Imt imt) {
      super(imt);
    }

    @Override
    String getRegion() {
      return Region.CASCADIA;
    }

    @Override
    String getBasin() {
      return Basin.PACNW_NONE;
    }
  }

  /*
   * Cascadia intraslab model
   */
  static class CascadiaSlab extends Slab {
    static final String NAME = Slab.NAME + " : Cascadia";

    CascadiaSlab(Imt imt) {
      super(imt);
    }

    @Override
    String getRegion() {
      return Region.CASCADIA;
    }

    @Override
    String getBasin() {
      return Basin.PACNW_NONE;
    }
  }

  /*
   * Cascadia interface model, Seattle basin
   */
  static class CascadiaSeattleInterface extends CascadiaInterface {
    static final String NAME = CascadiaInterface.NAME + " : Seattle";

    CascadiaSeattleInterface(Imt imt) {
      super(imt);
    }

    @Override
    String getBasin() {
      return Basin.SEATTLE;
    }
  }

  /*
   * Cascadia intraslab model, Seattle basin
   */
  static class CascadiaSeattleSlab extends CascadiaSlab {
    static final String NAME = CascadiaSlab.NAME + " : Seattle";

    CascadiaSeattleSlab(Imt imt) {
      super(imt);
    }

    @Override
    String getBasin() {
      return Basin.SEATTLE;
    }
  }

  /*
   * Cascadia interface model, Other PacNW basin
   */
  static class CascadiaPacNWInterface extends CascadiaInterface {
    static final String NAME = CascadiaInterface.NAME + " : PacNW";

    CascadiaPacNWInterface(Imt imt) {
      super(imt);
    }

    @Override
    String getBasin() {
      return Basin.PACNW;
    }
  }

  /*
   * Cascadia intraslab model, Other PacNW basin
   */
  static class CascadiaPacNWSlab extends CascadiaSlab {
    static final String NAME = CascadiaSlab.NAME + " : PacNW";

    CascadiaPacNWSlab(Imt imt) {
      super(imt);
    }

    @Override
    String getBasin() {
      return Basin.PACNW;
    }
  }

  /*
   * Alaska interface model
   */
  static class AlaskaInterface extends Interface {
    static final String NAME = Interface.NAME + " : Alaska";

    AlaskaInterface(Imt imt) {
      super(imt);
    }

    @Override
    String getRegion() {
      return Region.ALASKA;
    }
  }

  /*
   * Alaska intraslab model
   */
  static class AlaskaSlab extends Slab {
    static final String NAME = Slab.NAME + " : Alaska";

    AlaskaSlab(Imt imt) {
      super(imt);
    }

    @Override
    String getRegion() {
      return Region.ALASKA;
    }
  }

  /*
   * Aleutians interface model
   */
  static class AleutiansInterface extends Interface {
    static final String NAME = Interface.NAME + " : Aleutians";

    AleutiansInterface(Imt imt) {
      super(imt);
    }

    @Override
    String getRegion() {
      return Region.ALEUTIANS;
    }
  }

  /*
   * Aleutians intraslab model
   */
  static class AleutiansSlab extends Slab {
    static final String NAME = Slab.NAME + " : Aleutians";

    AleutiansSlab(Imt imt) {
      super(imt);
    }

    @Override
    String getRegion() {
      return Region.ALEUTIANS;
    }
  }

  public static void main(String[] args) {

  }

}
