package gov.usgs.earthquake.nshmp.internal.www;

import java.io.UnsupportedEncodingException;
import java.net.URI;
import java.net.URLDecoder;

import org.reactivestreams.Publisher;

import io.micronaut.context.annotation.Value;
import io.micronaut.core.type.MutableHeaders;
import io.micronaut.http.HttpHeaders;
import io.micronaut.http.HttpRequest;
import io.micronaut.http.MutableHttpResponse;
import io.micronaut.http.annotation.Filter;
import io.micronaut.http.filter.HttpServerFilter;
import io.micronaut.http.filter.ServerFilterChain;
import io.reactivex.Flowable;
import io.reactivex.schedulers.Schedulers;

/**
 * Custom NSHMP servlet implementation and URL helper class for Micronaut
 * services.
 *
 * <p>This class sets custom response headers and provides a helper class to
 * ensure serialized response URLs propagate the correct host and protocol from
 * requests on USGS servers and caches that may have been forwarded.
 *
 * @author U.S. Geological Survey
 */
@Filter("/**")
public class NshmpMicronautServlet implements HttpServerFilter {

  @Value("${micronaut.server.context-path:}")
  private String contextPath;

  /*
   * Set CORS headers and content type.
   *
   * Because NSHMP services may be called by both the USGS website, other
   * websites, and directly by 3rd party applications, responses generated by
   * direct requests will not have the necessary header information that would
   * be required by security protocols for web requests. This means that any
   * initial direct request will pollute intermediate caches with a response
   * that a browser will deem invalid.
   */
  @Override
  public Publisher<MutableHttpResponse<?>> doFilter(
      HttpRequest<?> request,
      ServerFilterChain chain) {
    return Flowable.just(chain).subscribeOn(Schedulers.io())
        .switchMap(bool -> chain.proceed(request))
        .doOnNext(res -> {
          MutableHeaders headers = res.getHeaders();
          headers.add("Access-Control-Allow-Origin", "*");
          headers.add("Access-Control-Allow-Methods", "*");
          headers.add("Access-Control-Allow-Headers", "accept,origin,authorization,content-type");
        });
  }

  public UrlHelper urlHelper(HttpRequest<?> request) {
    return new UrlHelper(request);
  }

  public class UrlHelper {

    public final String urlPrefix;
    public final String url;

    UrlHelper(HttpRequest<?> request) {
      HttpHeaders headers = request.getHeaders();
      URI uri = request.getUri();

      /*
       * Check custom header for a forwarded protocol so generated links can use
       * the same protocol and not cause mixed content errors.
       */
      String host = request.getServerName();
      String protocol = headers.get("X-FORWARDED-PROTO");
      if (protocol == null) {
        /* Not a forwarded request. Honor reported protocol and port. */
        protocol = uri.getScheme();
        host += ":" + request.getServerAddress().getPort();
      }

      String protocolHost = protocol == null ? host : protocol + "://" + host;

      /*
       * For convenience, store a url field with the (possibly updated) request
       * protocol
       */
      String url = uri.toString();
      try {
        url = protocolHost + URLDecoder.decode(uri.toString(), "UTF-8");
      } catch (UnsupportedEncodingException e) {
        // continue
      }

      this.urlPrefix = String.format("%s/%s", protocolHost, contextPath);
      this.url = url;
    }
  }

}
