package gov.usgs.earthquake.nshmp.model;

import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.base.Preconditions.checkState;
import static gov.usgs.earthquake.nshmp.model.SourceType.FAULT_CLUSTER;

import java.util.Iterator;
import java.util.List;
import java.util.function.Predicate;

import com.google.common.collect.Iterables;
import com.google.common.collect.Lists;

import gov.usgs.earthquake.nshmp.geo.Location;

/**
 * Container class for related {@link ClusterRuptureSet}s.
 *
 * @author U.S. Geological Survey
 * @see ClusterRuptureSet
 */
public class ClusterSourceSet extends AbstractSourceSet<ClusterSource> {

  private final List<ClusterSource> sources;

  private ClusterSourceSet(Builder builder) {
    super(builder);
    this.sources = builder.sources;
  }

  @Override
  public Iterator<ClusterSource> iterator() {
    return sources.iterator();
  }

  @Override
  public int size() {
    return sources.size();
  }

  @Override
  public SourceType type() {
    return FAULT_CLUSTER;
  }

  @Override
  public Predicate<ClusterSource> distanceFilter(
      final Location loc,
      final double distance) {

    return new Predicate<ClusterSource>() {

      private final Predicate<FaultSource> filter =
          new FaultSourceSet.DistanceFilter(loc, distance);

      @Override
      public boolean test(ClusterSource cs) {
        return Iterables.any(cs.faults, filter::test);
      }

      @Override
      public String toString() {
        return "ClusterSourceSet.DistanceFilter [ " + filter.toString() + " ]";
      }
    };
  }

  /* Single use builder */
  static class Builder extends AbstractSourceSet.Builder {

    private static final String ID = "ClusterSourceSet.Builder";

    private final List<ClusterSource> sources = Lists.newArrayList();

    Builder source(ClusterSource source) {
      sources.add(checkNotNull(source, "ClusterSource is null"));
      return this;
    }

    @Override
    void validateState(String buildId) {
      super.validateState(buildId);
      checkState(sources.size() > 0, "%s source list is empty", buildId);
    }

    ClusterSourceSet buildClusterSet() {
      validateState(ID);
      return new ClusterSourceSet(this);
    }
  }

}
