package gov.usgs.earthquake.nshmp.model;

import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.base.Preconditions.checkState;
import static gov.usgs.earthquake.nshmp.model.SourceType.FAULT;

import java.util.Iterator;
import java.util.List;
import java.util.function.Predicate;

import com.google.common.collect.Lists;

import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.geo.Locations;
import gov.usgs.earthquake.nshmp.tree.LogicGroup;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/**
 * Container class for related {@link FaultSource}s.
 *
 * @author U.S. Geological Survey
 * @see FaultSource
 */
public class FaultSourceSet extends AbstractSourceSet<FaultSource> {

  private final List<FaultSource> sources;

  /*
   * This will merge with SourceTree instance - currently are directly scaling
   * source total mfds by the branch weight to make existing calculators work,
   * but really want to update calculators to scale the individual RuptureSet
   * curves from SourceTree when combining.
   */
  private final LogicTree<FaultSource> sourceTree;

  private FaultSourceSet(Builder builder) {
    super(builder);
    this.sources = builder.sources;
    this.sourceTree = builder.sourceTree;
  }

  @Override
  public Iterator<FaultSource> iterator() {
    return sources.iterator();
  }

  @Override
  public int size() {
    return sources.size();
  }

  @Override
  public SourceType type() {
    return FAULT;
  }

  @Override
  public Predicate<FaultSource> distanceFilter(Location loc, double distance) {
    return new DistanceFilter(loc, distance);
  }

  /* Not inlined for use by cluster sources */
  static class DistanceFilter implements Predicate<FaultSource> {
    private final Predicate<Location> filter;

    DistanceFilter(Location loc, double distance) {
      filter = Locations.distanceFilter(loc, distance);
    }

    @Override
    public boolean test(FaultSource source) {
      return filter.test(source.feature.trace.first()) ||
          filter.test(source.feature.trace.last());
    }

    @Override
    public String toString() {
      return "FaultSourceSet.DistanceFilter[ " + filter.toString() + " ]";
    }
  }

  /* Single use builder. */
  static class Builder extends AbstractSourceSet.Builder {

    static final String ID = "FaultSourceSet.Builder";

    final List<FaultSource> sources = Lists.newArrayList();
    final LogicGroup.Builder<FaultSource> treeBuilder = LogicGroup.builder();
    LogicTree<FaultSource> sourceTree;

    Builder source(FaultSource source, double weight) {
      sources.add(checkNotNull(source, "FaultSource is null"));
      treeBuilder.addBranch(source.name(), source, weight);
      return this;
    }

    @Override
    void validateState(String buildId) {
      super.validateState(buildId);
    }

    FaultSourceSet buildFaultSet() {
      validateState(ID);
      // this is here because calls up from interface
      // will fail in this.validateState()
      checkState(sources.size() > 0, "%s source list is empty", ID);
      sourceTree = treeBuilder.build();
      return new FaultSourceSet(this);
    }
  }

}
