package gov.usgs.earthquake.nshmp.model;

import static com.google.common.base.Preconditions.checkState;
import static gov.usgs.earthquake.nshmp.model.Deserialize.MFD_TREE;
import static java.util.stream.Collectors.toUnmodifiableList;
import static java.util.stream.Collectors.toUnmodifiableMap;

import java.nio.file.Path;
import java.util.ArrayList;
import java.util.EnumMap;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.OptionalDouble;
import java.util.stream.Stream;

import gov.usgs.earthquake.nshmp.Maths;
import gov.usgs.earthquake.nshmp.data.DelimitedData;
import gov.usgs.earthquake.nshmp.data.DelimitedData.Record;
import gov.usgs.earthquake.nshmp.data.XyPoint;
import gov.usgs.earthquake.nshmp.fault.FocalMech;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.geo.json.Feature;
import gov.usgs.earthquake.nshmp.geo.json.Properties;
import gov.usgs.earthquake.nshmp.mfd.Mfd;
import gov.usgs.earthquake.nshmp.mfd.Mfd.Properties.GutenbergRichter;
import gov.usgs.earthquake.nshmp.mfd.Mfd.Type;
import gov.usgs.earthquake.nshmp.mfd.Mfds;
import gov.usgs.earthquake.nshmp.tree.Branch;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

class GridLoader {

  private GridLoader() {}

  /*
   * TODO Currently specialized flavors of GR are used for double-counting
   * processing but it would be preferable to do it when the model loads. This
   * would permit structural changes to a fault model without having to
   * regnerate a gridded source model. TODO note in documentation.
   */

  /*
   * Reduce location mfd-tree to total MFD. All MFDs should be GR* with
   * variations in mMax or GR-type and backed by common magnitude arrays that
   * are not copied when MFDs are combined.
   */

  // TODO deal with GR_TAPER mix in in WUS

  /* Grid csv file column header and property keys. */
  private static class Key {
    static final String LON = "lon";
    static final String LAT = "lat";
    static final String STRIKE = "strike";
    static final String RATE = "rate";
    static final String NAME = "name";
    static final String A_VALUE = "a";
    static final String MMAX = "mMax";
    static final String STRIKE_SLIP_WEIGHT = "ss_wt";
    static final String REVERSE_WEIGHT = "r_wt";
    static final String NORMAL_WEIGHT = "n_wt";
  }

  static List<SlabRuptureSet> createSlabRuptureSets(ModelData data) {
    GrLoader loader = new GrLoader(data);
    return loader.createSlabRuptureSets();
  }

  static List<GridRuptureSet> createGridRuptureSets(ModelData data) {

    Optional<SourceFeature.Grid> singleIncremental = GridLoader.isIncremental(data);

    if (singleIncremental.isPresent()) {
      IncrLoader loader = new IncrLoader(singleIncremental.orElseThrow(), data);
      return List.of(loader.createRuptureSet());
    } else {
      GrLoader loader = new GrLoader(data);
      return loader.createGridRuptureSets();
    }
  }

  /*
   * Return a single feature only if there is only one feature with a MFD logic
   * tree with one branch and the mfd type of that branch is INCR.
   */
  private static Optional<SourceFeature.Grid> isIncremental(ModelData data) {

    Map<String, SourceFeature.Grid> featureMap = data.gridFeatureMap().orElseThrow();

    if (featureMap.size() > 1) {
      return Optional.empty();
    }

    SourceFeature.Grid feature = featureMap.values().stream().findFirst().orElseThrow();
    LogicTree<Mfd.Properties> mfdTree = Deserialize.mfdTree(
        feature.source.properties(),
        data);

    if (mfdTree.size() > 1) {
      return Optional.empty();
    }

    Mfd.Type type = mfdTree.get(0).value().type();
    return (type == Mfd.Type.INCR)
        ? Optional.of(feature)
        : Optional.empty();
  }

  /* Handles single INCR MFD and single feature (UCERF3). */
  private static class IncrLoader {

    private final ModelData data;
    private final SourceFeature.Grid feature;
    private final Path ratesFile;
    FeatureData featureData;

    private IncrLoader(
        SourceFeature.Grid feature,
        ModelData data) {

      this.data = data;
      this.feature = feature;
      this.ratesFile = data.gridRateFile().orElseThrow();
      this.featureData = new FeatureData();
      processNodes();
      checkState(!featureData.focalMechMaps.isEmpty());
    }

    private GridRuptureSet createRuptureSet() {
      GridRuptureSet grs = GridRuptureSet.builder()
          .feature(feature)
          .modelData(data)
          .nodeData(
              featureData.locations,
              featureData.mfds,
              featureData.mfdsTree,
              featureData.focalMechMaps)
          .build();
      return grs;
    }

    private void processNodes() {

      DelimitedData ratefile = DelimitedData.comma(ratesFile);

      LogicTree<Mfd.Properties> mfdTree = Deserialize.mfdTree(
          feature.source.properties(),
          data);

      /*
       * TODO improve this (bad single-branch assumption): get only branch of
       * mfd logic tree
       */
      checkState(mfdTree.size() == 1);
      Branch<Mfd.Properties> propsBranch = mfdTree.get(0);
      Mfd.Builder mfdBuilder = propsBranch.value().toBuilder();

      try (Stream<Record> s = ratefile.records()) {
        s.forEach(record -> processRecord(record, mfdBuilder));
      }

      featureData.mfdsTree = LogicTree.singleton(
          feature.name,
          propsBranch.id(),
          featureData.mfds);
    }

    private void processRecord(Record record, Mfd.Builder builder) {

      Location loc = Location.create(
          record.getDouble(Key.LON),
          record.getDouble(Key.LAT));

      Map<FocalMech, Double> focalMechMap = createFocalMechMap(
          record.getDouble(Key.STRIKE_SLIP_WEIGHT),
          record.getDouble(Key.REVERSE_WEIGHT),
          record.getDouble(Key.NORMAL_WEIGHT));

      IncrRates incrRates = new IncrRates(record);
      builder.transform(incrRates::set);

      featureData.locations.add(loc);
      featureData.focalMechMaps.add(focalMechMap);
      featureData.mfds.add(builder.build());
    }

    private static class IncrRates {

      final Record record;

      IncrRates(Record record) {
        this.record = record;
      }

      void set(XyPoint mBin) {
        mBin.set(record.getDouble(Double.toString(mBin.x())));
      }
    }
  }

  private static Map<FocalMech, Double> createFocalMechMap(
      double ssWeight,
      double rWeight,
      double nWeight) {

    // TODO this should have validation of weight values and totals
    EnumMap<FocalMech, Double> fmMap = new EnumMap<>(FocalMech.class);
    fmMap.put(FocalMech.STRIKE_SLIP, ssWeight);
    fmMap.put(FocalMech.REVERSE, rWeight);
    fmMap.put(FocalMech.NORMAL, nWeight);
    return fmMap;
  }

  /* Handles logic trees of GR MFDs and multiple features. */
  private static class GrLoader {

    private final ModelData data;
    private final Map<String, SourceFeature.Grid> featureMap;
    private final Path ratesFile;
    private final Map<String, LogicTree<Mfd>> modelMfdTrees;
    private final Map<String, FeatureData> dataMap;

    private GrLoader(ModelData data) {

      this.data = data;
      this.featureMap = data.gridFeatureMap().orElseThrow();
      this.ratesFile = data.gridRateFile().orElseThrow();
      this.modelMfdTrees = createModelMfds();
      this.dataMap = featureMap.keySet().stream()
          .collect(toUnmodifiableMap(
              s -> s,
              s -> new FeatureData()));
      processNodes();
    }

    private List<GridRuptureSet> createGridRuptureSets() {
      List<GridRuptureSet> ruptureSets = new ArrayList<>();
      for (Entry<String, FeatureData> entry : dataMap.entrySet()) {
        SourceFeature.Grid feature = featureMap.get(entry.getKey());
        FeatureData featureData = dataMap.get(entry.getKey());
        GridRuptureSet grs = GridRuptureSet.builder()
            .feature(feature)
            .modelData(data)
            .nodeData(
                featureData.locations,
                featureData.mfds,
                featureData.mfdsTree,
                null)
            .build();
        ruptureSets.add(grs);
      }
      return ruptureSets;
    }

    private List<SlabRuptureSet> createSlabRuptureSets() {
      List<SlabRuptureSet> ruptureSets = new ArrayList<>();
      for (Entry<String, FeatureData> entry : dataMap.entrySet()) {
        SourceFeature.Grid feature = featureMap.get(entry.getKey());
        FeatureData featureData = dataMap.get(entry.getKey());
        SlabRuptureSet grs = SlabRuptureSet.builder()
            .feature(feature)
            .modelData(data)
            .nodeData(
                featureData.locations,
                featureData.mfds,
                featureData.mfdsTree)
            .build();
        ruptureSets.add(grs);
      }
      return ruptureSets;
    }

    private void processNodes() {
      DelimitedData data = DelimitedData.comma(ratesFile);
      /* May or may not have mMax column in rate file. */
      boolean mMaxPresent = data.columnKeys().contains(Key.MMAX);
      try (Stream<Record> s = data.records()) {
        s.forEach(record -> processRecord(record, mMaxPresent));
      }

      /* Create total MFDs and transpose mfd tree lists. */
      for (FeatureData featureData : dataMap.values()) {
        featureData.mfds = reduceMfds(featureData.mfdTrees);
        featureData.mfdsTree = MfdTrees.transpose(featureData.mfdTrees);
      }
    }

    private List<Mfd> reduceMfds(List<LogicTree<Mfd>> mfdTrees) {
      return mfdTrees.stream()
          .map(MfdTrees::reduceMfdTree)
          .collect(toUnmodifiableList());
    }

    // private boolean temp = true;

    private void processRecord(Record record, boolean mMaxPresent) {

      Location loc = Location.create(
          record.getDouble(Key.LON),
          record.getDouble(Key.LAT));
      double a = Math.log10(record.getDouble(Key.A_VALUE));
      OptionalDouble mMax = mMaxPresent
          ? record.getOptionalDouble(Key.MMAX)
          : OptionalDouble.empty();

      String gridId = zoneForLocation(loc);

      FeatureData featureData = dataMap.get(gridId);
      featureData.locations.add(loc);

      LogicTree<Mfd> modelMfds = modelMfdTrees.get(gridId);
      LogicTree<Mfd> nodeMfds = createNodeMfdTree(modelMfds, a, mMax);

      // if (temp) {
      // System.out.println(nodeMfds);
      // System.out.println(a + " " + mMax);
      // for (Branch<Mfd> branch : nodeMfds) {
      // System.out.println(branch.value().properties().getAsGr().mMin());
      // System.out.println(branch.value());
      // }
      // temp = false;
      // }
      featureData.mfdTrees.add(nodeMfds);
    }

    private LogicTree<Mfd> createNodeMfdTree(
        LogicTree<Mfd> modelTree,
        double a,
        OptionalDouble mMax) {

      LogicTree.Builder<Mfd> nodeMfdTree = LogicTree.builder(MFD_TREE);
      for (Branch<Mfd> branch : modelTree) {

        Mfd.Properties.GutenbergRichter grProps = branch.value().properties().getAsGr();
        double mMin = grProps.mMin() + grProps.Δm() * 0.5;
        double rate = Mfds.gutenbergRichterRate(a, grProps.b(), mMin);
        Mfd.Builder nodeMfd = Mfd.Builder.from(branch.value())
            .scaleToIncrementalRate(rate);
        mMax.ifPresent(m -> truncate(grProps, nodeMfd, m));
        nodeMfdTree.addBranch(
            branch.id(),
            nodeMfd.build(),
            branch.weight());
      }
      return nodeMfdTree.build();
    }

    /*
     * Check that each MFD logic tree consists fo the same type, has the same
     * mMin and Δm. This ensures we'll be able to create different mMax MFDs
     * with the same magnitude bins.
     */
    private Map<String, LogicTree<Mfd>> createModelMfds() {
      Map<String, LogicTree<Mfd>> modelMfds = new HashMap<>();
      for (Entry<String, SourceFeature.Grid> entry : featureMap.entrySet()) {
        SourceFeature.Grid grid = entry.getValue();
        LogicTree<Mfd.Properties> propsTree = Deserialize.mfdTree(
            grid.source.properties(),
            data);
        LogicTree<Mfd> mfdTree = MfdTrees.grPropsTreeToMmaxMfdTree(propsTree);
        modelMfds.put(entry.getKey(), mfdTree);
      }
      return Map.copyOf(modelMfds);
    }

    private String zoneForLocation(Location location) {
      return featureMap.values().stream()
          .filter(grid -> grid.contains(location))
          .map(grid -> grid.name)
          .findFirst()
          .orElseThrow();
    }
  }

  /*
   * Data container for feature polygon.
   *
   * Different fields are used by GrLoader and IncrLoader.
   */
  private static class FeatureData {

    /* node specific data */
    List<Location> locations = new ArrayList<>();
    List<LogicTree<Mfd>> mfdTrees = new ArrayList<>();
    List<Map<FocalMech, Double>> focalMechMaps = new ArrayList<>();

    /* transposed and combined data */
    LogicTree<List<Mfd>> mfdsTree;
    List<Mfd> mfds = new ArrayList<>();
  }

  /* Truncate special GR cases; e.g. WUS double counting. */
  private static void truncate(
      GutenbergRichter grProps,
      Mfd.Builder builder,
      double mMax) {
    Mfd.Type type = grProps.type();
    if (type == Type.GR_MMAX_GR) {
      builder.transform(xy -> zeroAboveM(xy, Math.min(mMax, 6.5)));
    } else if (type == Type.GR_MMAX_SINGLE) {
      /*
       * Keeps the magnitude bin just above mMax if mMax > m_gr + Δm/2. Rounding
       * to 5 decimal places ensures that mMax/Δm doesn't equal *.00...001,
       * which ceil() would take up to *.1
       *
       * We really should just be cutting out the uppermost bin with overlap but
       * this is consistent with what was being done in legacy fortran codes.
       */
      double Δm = grProps.Δm();
      double m = Math.ceil(Maths.round(mMax / Δm, 5)) * Δm;
      builder.transform(xy -> zeroAboveM(xy, m));
    }
  }

  static GridSourceSet createGrid(
      SourceConfig.Grid config,
      Feature feature,
      List<Location> locations,
      List<Mfd> mfds,
      LogicTree<List<Mfd>> mfdsTree,
      Optional<List<Map<FocalMech, Double>>> focalMechMaps,
      GmmSet gmms,
      double weight) {

    Properties props = feature.properties();

    GridSourceSet.Builder builder = new GridSourceSet.Builder();
    builder.name(props.getString(Key.NAME).orElseThrow())
        .id(feature.idAsInt().getAsInt())
        .weight(weight)
        .gmms(gmms);
    builder.gridConfig(config);
    props.getDouble(Key.STRIKE).ifPresent(builder::strike);

    // System.out.println(mfds.get(0).properties().type());
    builder.locations(locations, mfds, mfdsTree, focalMechMaps);

    return builder.build();
  }

  private static void zeroAboveM(XyPoint xy, double m) {
    if (xy.x() > m) {
      xy.set(0.0);
    }
  }

  /* Read a lon,lat,rate file into the supplied arrays */
  static void readZoneRateFile(
      Path csv,
      List<Location> locations,
      List<Double> rates) {

    try (Stream<Record> s = DelimitedData.comma(csv).records()) {
      s.forEach(record -> {
        locations.add(Location.create(
            record.getDouble(Key.LON),
            record.getDouble(Key.LAT)));
        rates.add(record.getDouble(Key.RATE));
      });
    }
  }

}
