package gov.usgs.earthquake.nshmp.model;

import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.base.Preconditions.checkState;

import java.util.List;
import java.util.Map;
import java.util.Optional;

import gov.usgs.earthquake.nshmp.fault.FocalMech;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.mfd.Mfd;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/*
 * Gridded seismicity rupture set implementation.
 *
 * @author U.S. Geological Survey
 */
class GridRuptureSet implements RuptureSet {

  private final SourceFeature.Grid feature;
  private final ModelData data;

  private final List<Location> locations;
  private final List<Mfd> mfds;
  private final LogicTree<List<Mfd>> mfdsTree;
  private final Optional<List<Map<FocalMech, Double>>> focalMechMaps;

  private GridSourceSet gss;

  /*
   * Enforce expectation that all MFDs are GR*
   */

  GridRuptureSet(Builder builder) {
    this.feature = builder.feature;
    this.data = builder.data;

    this.locations = builder.locations;
    this.mfds = builder.mfds;
    this.mfdsTree = builder.mfdsTree;
    this.focalMechMaps = builder.focalMechMaps;
  }

  // TODO declare this in RuptureSet? Source T issue
  GridSourceSet sourceSet(double weight) {
    if (gss == null) {
      gss = createSourceSet(weight);
    }

    // System.out.println("--- GSS ---");
    // for (PointSource ptSrc : gss) {
    // System.out.println(ptSrc);
    // for (Rupture rup : ptSrc) {
    // System.out.println(rup);
    // break;
    // }
    // }
    // System.out.println(gss.mechMaps);

    return gss;
  }

  private GridSourceSet createSourceSet(double weight) {
    SourceConfig.Grid config = data.sourceConfig().asGrid();
    GridSourceSet grid = GridLoader.createGrid(
        config,
        feature.source,
        locations,
        mfds,
        mfdsTree,
        focalMechMaps,
        data.gmms(),
        weight);

    return grid;
  }

  @Override
  public String name() {
    return feature.name;
  }

  @Override
  public int id() {
    return feature.id;
  }

  @Override
  public int size() {
    return locations.size();
  }

  @Override
  public SourceType type() {
    return SourceType.GRID;
  }

  @Override
  public Location location(Location site) {
    // TODO Auto-generated method stub
    return null;
  }

  static Builder builder() {
    return new Builder();
  }

  static final class Builder {

    private boolean built = false;

    private SourceFeature.Grid feature;
    private ModelData data;

    private List<Location> locations;
    private List<Mfd> mfds;
    private LogicTree<List<Mfd>> mfdsTree;

    private Optional<List<Map<FocalMech, Double>>> focalMechMaps = Optional.empty();

    Builder feature(SourceFeature.Grid feature) {
      this.feature = feature;
      return this;
    }

    Builder modelData(ModelData data) {
      this.data = data;
      return this;
    }

    Builder nodeData(
        List<Location> locations,
        List<Mfd> mfds,
        LogicTree<List<Mfd>> mfdsTree,
        List<Map<FocalMech, Double>> focalMechMaps) {

      // TODO check sizes

      this.locations = locations;
      this.mfds = mfds;
      this.mfdsTree = mfdsTree;
      this.focalMechMaps = Optional.ofNullable(focalMechMaps);
      return this;
    }

    private void validateAndInit(String label) {
      checkState(!built, "Single use builder");
      checkNotNull(feature, "%s feature", label);
      checkNotNull(data, "%s MFD reference data", label);

      checkNotNull(locations, "%s locations", label);
      checkNotNull(mfds, "%s total MFDs", label);
      checkNotNull(mfdsTree, "%s MFDs logic tree", label);

      built = true;
    }

    /* Create a new rupture set. */
    GridRuptureSet build() {
      validateAndInit("GridRuptureSet.Builder");
      return new GridRuptureSet(this);
    }
  }
}
