package gov.usgs.earthquake.nshmp.model;

import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.base.Preconditions.checkState;
import static gov.usgs.earthquake.nshmp.Text.NEWLINE;

import java.io.IOException;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import com.google.common.base.Strings;
import com.google.common.collect.ImmutableSetMultimap;
import com.google.common.collect.Maps;
import com.google.common.collect.SetMultimap;
import com.google.common.collect.Sets;
import com.google.common.collect.Streams;
import com.google.gson.annotations.SerializedName;

import gov.usgs.earthquake.nshmp.Maths;
import gov.usgs.earthquake.nshmp.calc.CalcConfig;
import gov.usgs.earthquake.nshmp.gmm.Gmm;
import gov.usgs.earthquake.nshmp.gmm.GroundMotionModel;
import gov.usgs.earthquake.nshmp.gmm.NehrpSiteClass;
import gov.usgs.earthquake.nshmp.model.SourceTree.Leaf;
import gov.usgs.earthquake.nshmp.tree.Branch;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/**
 * A {@code HazardModel} is the top-level wrapper for earthquake {@link Source}
 * definitions and attendant {@link GroundMotionModel}s used in probabilisitic
 * seismic hazard analyses (PSHAs) and related calculations. A
 * {@code HazardModel} contains of a number of {@link SourceSet}s that define
 * logical groupings of sources by {@link SourceType}. Each {@code SourceSet}
 * carries with it references to the {@code GroundMotionModel}s and associated
 * weights to be used when evaluating hazard.
 *
 * @author U.S. Geological Survey
 * @see Source
 * @see SourceSet
 * @see SourceType
 */
public final class HazardModel implements Iterable<SourceSet<? extends Source>> {

  // TODO refactor to 'Model'

  // TODO Is current approach in SYstemSourceSet for precomputing distances
  // correct? Double check that the individual sections are scaled by aseis
  // not after once built into ruptures.

  // TODO UCERF3 xml (indexedFaultSource) needs to have aftershock correction
  // either imposed in file (better) of be imlemented in Parser
  // or Loader (worse)

  // TODO are any M=6.0 finite source representation cutoffs being used in
  // 2014 fortran

  // low priority
  // TODO test low rate shortcut in FaultSource
  // "if (rate < 1e-14) continue; // shortcut low rates"

  // TODO need to revisit the application of uncertainty when minM < 6.5
  // e.g. 809a Pine Valley graben in orwa.c.in

  private final String name;
  private final Map<NehrpSiteClass, Double> siteClassMap;
  private final SetMultimap<SourceType, SourceSet<? extends Source>> sourceSetMap;
  private final CalcConfig config;

  private HazardModel(Builder builder) {
    this.name = builder.info.name;
    this.siteClassMap = Maps.immutableEnumMap(builder.info.siteClassMap);
    this.config = builder.config;
    this.sourceSetMap = builder.sourceSetMap;
  }

  /**
   * Load a {@code HazardModel} from a directory specified by the supplied
   * {@code path}.
   *
   * <p>For more information on a HazardModel directory and file structure, see
   * the <a
   * href="https://github.com/usgs/nshmp-haz/wiki/Earthquake-Source-Models"
   * target="_top">nshmp-haz wiki</a>.
   *
   * <p><b>Notes:</b> HazardModel loading is not thread safe. There are also a
   * wide variety of exceptions that may be encountered when loading a model.
   * Exceptions are generally logged and propagated back to calling application
   * to handle termination.
   *
   * @param path to {@code HazardModel} directory
   * @return a newly instantiated {@code HazardModel}
   */
  public static HazardModel load(Path path) throws IOException {
    return ModelLoader.load(path);
  }

  /**
   * The number of {@code SourceSet}s in this {@code HazardModel}.
   */
  public int size() {
    return sourceSetMap.size();
  }

  @Override
  public Iterator<SourceSet<? extends Source>> iterator() {
    return sourceSetMap.values().iterator();
  }

  /**
   * The name of the model.
   */
  public String name() {
    return name;
  }

  /**
   * Return the default calculation configuration. This may be overridden.
   */
  public CalcConfig config() {
    return config;
  }

  /**
   * The set of {@code Gmm}s used by this model.
   */
  public Set<Gmm> gmms() {
    return Streams.stream(this)
        .flatMap(ss -> ss.groundMotionModels().gmms().stream())
        .collect(Sets.toImmutableEnumSet());
  }

  /**
   * The set of {@code SourceType}s included in this model.
   */
  public Set<SourceType> types() {
    return sourceSetMap.keySet();
  }

  public Map<NehrpSiteClass, Double> siteClasses() {
    return siteClassMap;
  }

  @Override
  public String toString() {
    StringBuilder sb = new StringBuilder("HazardModel: ");
    sb.append(name).append(NEWLINE);
    for (SourceSet<? extends Source> sourceSet : this) {
      String typeStr = "  " + sourceSet.type() + " Source Set";
      sb.append(Strings.padEnd(typeStr, 26, ' '));
      sb.append(sourceSet);
      sb.append(NEWLINE);
    }
    return sb.toString();
  }

  /**
   * Returns a new {@link Builder}.
   */
  static Builder builder() {
    return new Builder();
  }

  static class Builder {

    static final String ID = "HazardModel.Builder";
    boolean built = false;

    // ImmutableSetMultimap.Builder preserves value addition order
    private ImmutableSetMultimap.Builder<SourceType, SourceSet<? extends Source>> sourceMapBuilder;
    private SetMultimap<SourceType, SourceSet<? extends Source>> sourceSetMap;
    private Info info;
    private CalcConfig config;

    // TODO not used but need to be
    private List<SourceTree> trees = new ArrayList<>();

    private Builder() {
      sourceMapBuilder = ImmutableSetMultimap.builder();
    }

    Builder info(Info info) {
      this.info = checkNotNull(info);
      return this;
    }

    Builder config(CalcConfig config) {
      this.config = checkNotNull(config);
      return this;
    }

    private Builder addSourceTree(SourceTree tree) {
      // System.out.println("model: source tree added: " + tree.name());
      trees.add(tree);
      addSourceSetsFromTree(tree);
      return this;
    }

    Builder addSourceTrees(List<SourceTree> trees) {
      trees.forEach(this::addSourceTree);
      return this;
    }

    private Builder addSourceSet(SourceSet<? extends Source> source) {
      // System.out.println(source);
      sourceMapBuilder.put(source.type(), source);
      return this;
    }

    private void validateState(String mssgID) {
      checkState(!built, "This %s instance as already been used", mssgID);
      checkState(info != null, "%s info not set", mssgID);
      checkState(config != null, "%s config not set", mssgID);
      checkState(sourceSetMap.size() > 0, "%s has no source sets", mssgID);
      built = true;
    }

    HazardModel build() {
      sourceSetMap = sourceMapBuilder.build();
      validateState(ID);
      return new HazardModel(this);
    }

    private void addSourceSetsFromTree(SourceTree tree) {
      switch (tree.type()) {
        case ZONE:
          zoneSourceSetsFromTree(tree);
          break;
        case FAULT:
        case FAULT_CLUSTER:
        case FAULT_SYSTEM:
          faultSourceSetsFromTree(tree);
          break;
        case GRID:
          gridSourceSetsFromTree(tree);
          break;
        case INTRASLAB:
          slabSourceSetsFromTree(tree);
          break;
        case INTERFACE:
          interfaceSourceSetsFromTree(tree);
          break;
        default:
          throw new UnsupportedOperationException();
      }
    }

    /*
     * TODO We really should be able to collapse cluster rate tree as well;
     * never really thought about this because the rate tree was formerly
     * isolated in individual\source files, but now we've split the top level
     * source-tree branches on geometry/position
     */
    private void faultSourceSetsFromTree(SourceTree tree) {

      /* SYSTEM: each leaf is a SystemSourceSet */
      if (tree.type() == SourceType.FAULT_SYSTEM) {
        for (Leaf leaf : tree.branchMap().keySet()) {
          RuptureSet rs = leaf.ruptureSet;
          double branchWeight = tree.branchWeights().get(leaf);
          SystemRuptureSet srs = (SystemRuptureSet) rs;
          SystemSourceSet sss = srs.createSourceSet(branchWeight);
          addSourceSet(sss);
        }
        return;
      }

      /* FAULT: each leaf is a FaultSource in a FaultSourceSet */
      // System.out.println(tree.singleType());
      if (tree.singleType()) {

        FaultSourceSet.Builder fss = new FaultSourceSet.Builder();
        fss.name(tree.name());
        fss.weight(1.0);
        fss.gmms(tree.gmms());

        for (Leaf leaf : tree.branchMap().keySet()) {
          RuptureSet rs = leaf.ruptureSet;
          // System.out.println("Flt to FaultSrcSet: " + rs.type() + " " +
          // rs.name());

          double leafWeight = tree.branchWeights().get(leaf);
          FaultRuptureSet frs = (FaultRuptureSet) rs;

          /*
           * TODO we're overwriting the id with the id of each rupture set leaf
           * for dip variants this is ok, but there are many other sources where
           * this WILL NOT hold; rupture-sets commonly have unique IDs that are
           * distinct from the (possibly multiple) underlying fault sections.
           */
          fss.id(frs.id());

          // TODO leaf weight is being applied to total MFD
          FaultSource faultSource = faultRuptureSetToSource(frs, leafWeight);
          // System.out.println("FS size: " + faultSource.size());

          if (faultSource.size() > 0) {
            fss.source(faultSource, leafWeight);
          } else {
            System.out.println("      No source ruptures: " + faultSource.name());
          }
        }
        if (fss.sources.size() > 0) {
          addSourceSet(fss.buildFaultSet());
        } else {
          System.out.println("      Empty Source Set: " + fss.name);
        }
        return;
      }

      /* Otherwise types are mixed, e.g. New Madrid, Wasatch */
      for (Leaf leaf : tree.branchMap().keySet()) {

        RuptureSet rs = leaf.ruptureSet;

        double branchWeight = tree.branchWeights().get(leaf);

        if (leaf.ruptureSet.type() == SourceType.FAULT_CLUSTER) {

          // System.out.println("Flt to ClusterSrcSet: " + rs.type() + " " +
          // rs.name());

          ClusterRuptureSet crs = (ClusterRuptureSet) rs;

          ClusterSourceSet.Builder css = new ClusterSourceSet.Builder();
          css.name(crs.name);
          css.id(crs.id);
          css.gmms(tree.gmms());
          css.weight(branchWeight); // NM geometry branch

          // TODO consider ability to collabse cluster rate-tree.
          LogicTree<Double> rateTree = crs.data.rateTree().orElseThrow();

          for (Branch<Double> rateBranch : rateTree) {

            double returnPeriod = rateBranch.value();
            double rate = (returnPeriod <= 0.0)
                ? 0.0
                : Maths.roundToDigits(1.0 / returnPeriod, 4);

            String clusterLabel = crs.name + " : " + ((int) returnPeriod) + "-yr";
            ClusterSource.Builder csBuilder = new ClusterSource.Builder()
                .name(clusterLabel)
                .id(crs.id)
                .rate(rate); // cluster rate

            FaultSourceSet.Builder fss = new FaultSourceSet.Builder();
            fss.name(clusterLabel);
            fss.id(crs.id);
            fss.gmms(tree.gmms());

            // The weight of the fss gets called by ClusterSource.weight()
            // and is used in HazardCurveSet.Builder.addCurves()
            fss.weight(rateBranch.weight());

            for (FaultRuptureSet frs : crs.faultRuptureSets) {
              fss.source(faultRuptureSetToSource(frs, 1.0), 1.0);
            }
            csBuilder.faults(fss.buildFaultSet());
            css.source(csBuilder.buildClusterSource());

          }

          addSourceSet(css.buildClusterSet());

        }
      }
    }

    private FaultSource faultRuptureSetToSource(
        FaultRuptureSet frs,
        double leafWeight) {

      return new FaultSource.Builder()
          .feature(frs.feature)
          .config(frs.data.sourceConfig().asFault())
          .mfdTree(frs.mfdTree)
          .mfdTotal(frs.mfdTotal)
          .branchWeight(leafWeight)
          .build();
    }

    private void interfaceSourceSetsFromTree(SourceTree tree) {

      InterfaceSourceSet.Builder builder = new InterfaceSourceSet.Builder();
      builder.name(tree.name());
      builder.id(-1);
      builder.weight(1.0);
      builder.gmms(tree.gmms());

      for (Leaf leaf : tree.branchMap().keySet()) {
        double leafWeight = tree.branchWeights().get(leaf);
        InterfaceRuptureSet irs = (InterfaceRuptureSet) leaf.ruptureSet;
        InterfaceSource is = interfaceRuptureSetToSource(irs, leafWeight);
        builder.source(is, leafWeight);
      }
      addSourceSet(builder.buildSubductionSet());
    }

    private InterfaceSource interfaceRuptureSetToSource(
        InterfaceRuptureSet irs,
        double weight) {

      return new InterfaceSource.Builder()
          .feature(irs.feature)
          .config(irs.data.sourceConfig().asInterface())
          .mfdTree(irs.mfdTree)
          .mfdTotal(irs.mfdTotal)
          .branchWeight(weight)
          .build();
    }

    private void zoneSourceSetsFromTree(SourceTree tree) {
      for (Leaf leaf : tree.branchMap().keySet()) {
        double leafWeight = tree.branchWeights().get(leaf);
        ZoneRuptureSet zrs = (ZoneRuptureSet) leaf.ruptureSet;
        addSourceSet(zrs.sourceSet(leafWeight));
      }
    }

    private void slabSourceSetsFromTree(SourceTree tree) {
      for (Leaf leaf : tree.branchMap().keySet()) {
        double leafWeight = tree.branchWeights().get(leaf);
        SlabRuptureSet srs = (SlabRuptureSet) leaf.ruptureSet;
        addSourceSet(srs.sourceSet(leafWeight));
      }
    }

    private void gridSourceSetsFromTree(SourceTree tree) {
      for (Leaf leaf : tree.branchMap().keySet()) {
        double leafWeight = tree.branchWeights().get(leaf);
        GridRuptureSet grs = (GridRuptureSet) leaf.ruptureSet;
        addSourceSet(grs.sourceSet(leafWeight));
      }
    }

  }

  static class Info {

    final String name;
    @SerializedName("site-class-vs30")
    final Map<NehrpSiteClass, Double> siteClassMap;

    Info(String name, Map<NehrpSiteClass, Double> siteClassMap) {
      this.name = name;
      this.siteClassMap = siteClassMap;
    }
  }

}
