package gov.usgs.earthquake.nshmp.model;

import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.base.Preconditions.checkState;
import static gov.usgs.earthquake.nshmp.model.SourceType.INTERFACE;

import java.util.Iterator;
import java.util.List;
import java.util.function.Predicate;

import com.google.common.collect.Lists;

import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.geo.Locations;
import gov.usgs.earthquake.nshmp.tree.LogicGroup;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/**
 * Container class for groups of related {@link InterfaceSource}s.
 *
 * @author U.S. Geological Survey
 * @see InterfaceSource
 */
public class InterfaceSourceSet extends AbstractSourceSet<InterfaceSource> {

  private final List<InterfaceSource> sources;

  /*
   * TODO source group; Cascadia logic tree does not sum to one
   *
   * This will merge with SourceTree instance - currently are directly scaling
   * source total mfds by the branch weight to make existing calculators work,
   * but really want to update calculators to scale the individual RuptureSet
   * curves from SourceTree when combining.
   */

  /* TODO Not used. Source weight is built into source MFDs */
  private final LogicTree<InterfaceSource> sourceTree;

  private InterfaceSourceSet(Builder builder) {
    super(builder);
    this.sources = builder.sources;
    this.sourceTree = builder.sourceTree;

    // TODO clean
    // for (InterfaceSource src : sources) {
    // src.mfds().stream().map(Mfd::data).forEach(System.out::println);
    // }
  }

  @Override
  public Iterator<InterfaceSource> iterator() {
    return sources.iterator();
  }

  @Override
  public int size() {
    return sources.size();
  }

  @Override
  public SourceType type() {
    return INTERFACE;
  }

  @Override
  public Predicate<InterfaceSource> distanceFilter(final Location loc,
      final double distance) {

    return new Predicate<InterfaceSource>() {
      private Predicate<Location> filter = Locations.distanceFilter(loc, distance);

      @Override
      public boolean test(InterfaceSource source) {
        return filter.test(source.upperTrace.first()) ||
            filter.test(source.upperTrace.last()) ||
            filter.test(source.lowerTrace.first()) ||
            filter.test(source.lowerTrace.last());
      }

      @Override
      public String toString() {
        return "InterfaceSourceSet.DistanceFilter[ " + filter.toString() + " ]";
      }
    };
  }

  /* Single use builder. */
  static class Builder extends FaultSourceSet.Builder {

    static final String ID = "InterfaceSourceSet.Builder";

    final List<InterfaceSource> sources = Lists.newArrayList();
    final LogicGroup.Builder<InterfaceSource> treeBuilder = LogicGroup.builder();
    LogicTree<InterfaceSource> sourceTree;

    Builder source(InterfaceSource source, double weight) {
      sources.add(checkNotNull(source, "InterfaceSource is null"));
      treeBuilder.addBranch(source.name(), source, weight);
      return this;
    }

    InterfaceSourceSet buildSubductionSet() {
      validateState(ID);
      sourceTree = treeBuilder.build();
      checkState(sources.size() > 0, "%s source list is empty", id);
      return new InterfaceSourceSet(this);
    }
  }

}
