package gov.usgs.earthquake.nshmp.model;

import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.base.Preconditions.checkState;
import static gov.usgs.earthquake.nshmp.model.DipSlipModel.FIXED;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.FAULT_SOURCES;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.GRID_SOURCES;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.ZONE_SOURCES;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.checkGridDataDirectory;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.checkSourceTree;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.checkTectonicSetting;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.readFaultClusterSet;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.readFaultConfig;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.readFaultFeatures;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.readFaultRateTree;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.readFaultRuptureSet;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.readGmms;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.readGridConfig;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.readGridRateTree;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.readInterfaceConfig;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.readInterfaceFeatures;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.readInterfaceRuptureSet;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.readMfdConfig;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.readMfdMap;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.readModelInfo;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.readRateTree;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.readSlabConfig;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.readSourceTree;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.readSystemRuptureSet;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.readZoneConfig;
import static gov.usgs.earthquake.nshmp.model.SourceType.FAULT;
import static gov.usgs.earthquake.nshmp.model.SourceType.FAULT_SYSTEM;
import static gov.usgs.earthquake.nshmp.model.SourceType.GRID;
import static gov.usgs.earthquake.nshmp.model.SourceType.INTERFACE;
import static gov.usgs.earthquake.nshmp.model.SourceType.INTRASLAB;
import static gov.usgs.earthquake.nshmp.model.SourceType.ZONE;
import static java.nio.file.Files.newDirectoryStream;

import java.io.IOException;
import java.math.RoundingMode;
import java.nio.file.DirectoryStream;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.google.common.collect.Iterables;
import com.google.gson.JsonElement;

import gov.usgs.earthquake.nshmp.Maths;
import gov.usgs.earthquake.nshmp.fault.FocalMech;
import gov.usgs.earthquake.nshmp.geo.json.Feature;
import gov.usgs.earthquake.nshmp.geo.json.Properties;
import gov.usgs.earthquake.nshmp.mfd.Mfd;
import gov.usgs.earthquake.nshmp.model.SourceFeature.Key;
import gov.usgs.earthquake.nshmp.tree.Branch;
import gov.usgs.earthquake.nshmp.tree.LogicGroup;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/**
 * Earthquake source model loading class.
 *
 * Although source types are broadly similar, differences in model structure
 * necessitate source type specific loaders. Models are loaded recursively and
 * much of the heavy lifting of checking for the presence of, reading, and
 * deserializing are handled by the factory classes ModelFiles and Deserialize.
 *
 * @author U.S. Geological Survey
 */
abstract class ModelLoader {

  public static void main(String[] args) {
    Path testModel = Paths.get("../nshm-conus-2018");
    HazardModel model = ModelLoader.load(testModel);
    System.out.println(model);
  }

  private static Logger log;

  static {
    log = Logger.getLogger(ModelLoader.class.getName());
  }

  static HazardModel load(Path model) {

    try {
      System.out.println(model);

      HazardModel.Builder builder = HazardModel.builder()
          .info(readModelInfo(model).orElseThrow());

      /* Look for optional calc config. */
      ModelFiles.readCalcConfig(model, builder::config);

      /* Process tectonic setting source directories. */
      ModelFiles.readDirectories(
          model,
          ModelLoader::loadTectonicSetting,
          builder::addSourceTrees);

      return builder.build();

    } catch (Exception e) {
      String pathString = (model != null)
          ? model.toAbsolutePath().normalize().toString()
          : "null";
      log.log(Level.SEVERE, "Error loading model: " + pathString, e);
      // log.severe("Error loading model: " + path);
      throw e;
    }
  }

  // TODO clean - just here so static imports don't get deleted
  private static final String[] tmp = new String[] { FAULT_SOURCES, GRID_SOURCES, ZONE_SOURCES };

  /* Process teconic setting directories. */
  private static List<SourceTree> loadTectonicSetting(Path dir) {
    Optional<TectonicSetting> setting = checkTectonicSetting(dir);
    /* Return empty tree list for non-matching diectories. */
    if (setting.isEmpty()) {
      return List.of();
    }
    switch (setting.orElseThrow()) {
      case ACTIVE_CRUST:
      case STABLE_CRUST:
        return crustalSources(dir);
      case SUBDUCTION_INTERFACE:
        return interfaceSources(dir);
      case SUBDUCTION_SLAB:
        return slabSources(dir);
      default: // TODO this can go away if all settings present
        return List.of();
      // throw new UnsupportedOperationException("Unsupported source type");
    }
  }

  private static List<SourceTree> crustalSources(Path dir) {

    List<SourceTree> trees = new ArrayList<>();

    GmmSet gmms = ModelFiles.readGmms(dir).orElseThrow();

    try (DirectoryStream<Path> paths = newDirectoryStream(dir)) {
      for (Path path : paths) {
        if (!Files.isDirectory(path)) continue;
        String filename = checkNotNull(path.getFileName()).toString();
        switch (filename) {
          case FAULT_SOURCES:
            trees.addAll(faultSources(path, gmms));
            break;
          case GRID_SOURCES:
            trees.addAll(gridSources(path, gmms));
            break;
          case ZONE_SOURCES:
            trees.addAll(zoneSources(path, gmms));
            break;
          // default:
          // throw new UnsupportedOperationException("Unsupported source type: "
          // + filename);
        }
      }
    } catch (IOException ioe) {
      throw new RuntimeException(ioe);
    }
    return trees;
  }

  /* --------------- ↑ static - abstract ↓ --------------- */

  final Path root;

  /*
   * TODO In process*Branch methods we are exiting for null ("do nothing")
   * branches (so not adding them to source tree) and therefore can't sample
   */
  private ModelLoader(Path root) {
    this.root = root;
  }

  /* Load source type specific config, if present. */
  abstract void loadSourceConfig(Path dir, ModelData data);

  /* Load source type specific rate tree, if present. */
  abstract void loadRateTree(Path dir, ModelData data);

  /* Load single (branch) type specific source. */
  abstract SourceTree loadSingleSource(
      Path geojson,
      ModelData data);

  /* Load source type specific tree. */
  abstract SourceTree loadSourceTree(
      Path dir,
      ModelData data);

  /* Source type independent recursive directory processor. */
  List<SourceTree> loadSourceDirectory(
      Path dir,
      ModelData dataIn) {

    /* Update possible overrides in copy of support class. */
    ModelData data = ModelData.copyOf(dataIn);
    readGmms(dir).ifPresent(data::gmms);
    readMfdConfig(dir).ifPresent(data::mfdConfig);
    readMfdMap(dir).ifPresent(data::mfdMap);
    loadSourceConfig(dir, data);
    loadRateTree(dir, data);

    /* (1) Process source tree, if present. */
    if (checkSourceTree(dir).isPresent()) {
      return List.of(loadSourceTree(dir, data));
    }

    List<SourceTree> trees = new ArrayList<>();

    /* (2) Process any GeoJSON source files. */
    ModelFiles.readGeoJsons(dir, data, this::loadSingleSource, trees::add);

    /* (3) Lastly, recurse into any nested directories. */
    ModelFiles.readNestedDirectories(dir, data, this::loadSourceDirectory, trees::addAll);

    return trees;
  }

  private static void logRoot(Path root, SourceType type) {
    int end = root.getNameCount();
    int start = (type == INTRASLAB || type == INTERFACE) ? end - 1 : end - 2;
    System.out.println(root.subpath(start, end));
  }

  /* ------------ ↑ abstract - implementations ↓ ------------ */

  static List<SourceTree> faultSources(Path dir, GmmSet gmms) {
    SourceConfig srcConfig = readFaultConfig(dir).orElseThrow();
    ModelData data = ModelData.create(gmms, srcConfig);
    return new Fault(dir).loadSourceDirectory(dir, data);
  }

  private static class Fault extends ModelLoader {

    Fault(Path root) {
      super(root);
      logRoot(root, FAULT);
    }

    @Override
    void loadSourceConfig(Path dir, ModelData data) {
      readFaultConfig(dir).ifPresent(data::sourceConfig);
    }

    @Override
    void loadRateTree(Path dir, ModelData data) {
      readFaultRateTree(dir).ifPresent(data::faultRateTree);
    }

    @Override
    SourceTree loadSingleSource(
        Path geojson,
        ModelData data) {

      /* Ensure singleton mfd-map. */
      checkState(data.mfdMap().orElseThrow().size() == 1);

      SourceFeature.NshmFault feature = SourceFeature.newNshmFault(geojson);

      SourceTree.Builder treeBuilder = SourceTree.builder()
          .name(feature.name)
          .type(FAULT)
          .gmms(data.gmms());

      System.out.println(" source: " + root.relativize(geojson));

      return loadSingleSource(treeBuilder, Optional.empty(), feature, data).build();

      //
      // /* Possibly split on normal fault dip variants. */
      // SourceConfig.Fault config = data.sourceConfig().asFault();
      // boolean isNormal = FocalMech.fromRake(feature.rake) ==
      // FocalMech.NORMAL;
      //
      // if (config.dipTree.isPresent() && isNormal) {
      // System.out.println("dip tree: " + root.relativize(geojson));
      // // SourceTree.Builder treeBuilder = SourceTree.builder()
      // // .name(feature.name)
      // // .type(FAULT)
      // // .gmms(data.gmms());
      //
      // return loadNormalDipTree(treeBuilder, Optional.empty(), feature, data)
      // .build();
      // }
      //
      // System.out.println(" source: " + root.relativize(geojson));
      //
      // /*
      // * Create new feature with updated properties. Magnitude may be present
      // in
      // * properties for regional, historic, or other overrides.
      // */
      // // double Mw = feature.magnitude.isPresent()
      // // ? feature.magnitude.orElseThrow()
      // // : FaultRuptureSet.magnitudeWc94(feature.length);
      // // double Δz = feature.lowerDepth - feature.upperDepth;
      // // double width = computeWidth(Δz, feature.dip);
      // // Feature nshmFeature = Feature.copyOf(feature.source)
      // // .properties(Properties.fromFeature(feature.source)
      // // .put(Key.MAGNITUDE, Mw)
      // // .put(Key.WIDTH, width)
      // // .build())
      // // .build();
      // // feature = new SourceFeature.NshmFault(nshmFeature);
      //
      // // feature = updateFeature(feature);
      //
      // LogicTree<Path> root = LogicTree.singleton(
      // Deserialize.SOURCE_TREE,
      // feature.name,
      // Path.of(feature.name));
      //
      // treeBuilder.root(root);
      //
      // // SourceTree.Builder treeBuilder = SourceTree.builder()
      // // .name(feature.name)
      // // .type(FAULT)
      // // .gmms(data.gmms())
      // // .root(root);
      //
      // // RuptureSet ruptureSet = createSingleRuptureSet(data, feature);
      // //
      // // treeBuilder.addLeaf(root.get(0), ruptureSet);
      // //
      // // loadSingleRuptureSet(
      // // root.get(0),
      // // treeBuilder,
      // // data,
      // // feature);
      //
      // return loadSingle(treeBuilder, root.get(0), data, feature).build();
      //
      // // return treeBuilder.build();
    }

    private SourceTree.Builder loadSingleSource(
        SourceTree.Builder treeBuilder,
        Optional<Branch<Path>> branch,
        SourceFeature.NshmFault feature,
        ModelData data) {

      /* Possibly split on normal fault dip variants. */
      SourceConfig.Fault config = data.sourceConfig().asFault();
      boolean isNormal = FocalMech.fromRake(feature.rake) == FocalMech.NORMAL;

      if (config.dipTree.isPresent() && isNormal) {
        return loadNormalDipTree(treeBuilder, branch, feature, data);
      }

      return loadSingle(treeBuilder, branch, data, feature);
    }

    private SourceTree.Builder loadSingle(
        SourceTree.Builder treeBuilder,
        Optional<Branch<Path>> branch,
        ModelData data,
        SourceFeature.NshmFault feature) {

      /* If supplied branch is empty, we need to set a singleton tree root. */
      if (branch.isEmpty()) {
        LogicTree<Path> root = LogicTree.singleton(
            Deserialize.SOURCE_TREE,
            feature.name,
            Path.of(feature.name));
        treeBuilder.root(root);
        branch = Optional.of(root.get(0));
      }

      double Mw = feature.magnitude.isPresent()
          ? feature.magnitude.orElseThrow()
          : FaultRuptureSet.magnitudeWc94(feature.length);
      double Δz = feature.lowerDepth - feature.upperDepth;
      double width = computeWidth(Δz, feature.dip);
      Feature newFeature = Feature.copyOf(feature.source)
          .properties(Properties.fromFeature(feature.source)
              .put(Key.MAGNITUDE, Mw)
              .put(Key.WIDTH, width)
              .build())
          .build();
      feature = new SourceFeature.NshmFault(newFeature);

      RuptureSet ruptureSet = createSingleRuptureSet(data, feature);
      treeBuilder.addLeaf(branch.orElseThrow(), ruptureSet);
      return treeBuilder;
    }

    private SourceTree.Builder loadNormalDipTree(
        SourceTree.Builder treeBuilder,
        Optional<Branch<Path>> branch, // if absent, set tree root
        SourceFeature.NshmFault feature,
        ModelData data) {

      System.out.println("dip tree: ");

      SourceConfig.Fault config = data.sourceConfig().asFault();
      LogicTree<Double> dipTree = config.dipTree.orElseThrow();
      LogicTree<Path> dipSourceTree = MfdTrees.toPathTree(dipTree);
      DipSlipModel dipSlipModel = config.dipSlipModel.orElseThrow();
      // SourceTree.Builder treeBuilder = SourceTree.builder()
      // .name(feature.name)
      // .type(FAULT)
      // .gmms(data.gmms())
      // .root(dipSourceTree);

      if (branch.isPresent()) {
        treeBuilder.addBranches(branch.orElseThrow(), dipSourceTree);
      } else {
        treeBuilder.root(dipSourceTree);
      }

      /*
       * Create new features with updated properties. Magnitude may be present
       * in properties for regional, historic, or other overrides.
       */
      double Mw = feature.magnitude.isPresent()
          ? feature.magnitude.orElseThrow()
          : FaultRuptureSet.magnitudeWc94(feature.length);
      for (int i = 0; i < dipTree.size(); i++) {
        Branch<Double> dipBranch = dipTree.get(i);
        int branchDip = (int) (feature.dip + dipBranch.value());
        int rateDip = (int) ((dipSlipModel == FIXED) ? feature.dip : branchDip);
        String branchName = String.format("%s [%s°]", feature.name, branchDip);
        double Δz = feature.lowerDepth - feature.upperDepth;
        double width = computeWidth(Δz, branchDip);
        Feature dipFeature = Feature.copyOf(feature.source)
            .properties(Properties.fromFeature(feature.source)
                .put(Key.NAME, branchName)
                .put(Key.DIP, branchDip)
                .put(Key.RATE_DIP, rateDip)
                .put(Key.MAGNITUDE, Mw)
                .put(Key.WIDTH, width)
                .build())
            .build();

        RuptureSet ruptureSet = createSingleRuptureSet(
            data,
            new SourceFeature.NshmFault(dipFeature));

        treeBuilder.addLeaf(dipSourceTree.get(i), ruptureSet);

        // loadSingleRuptureSet(
        // dipSourceTree.get(i),
        // treeBuilder,
        // data,
        // new SourceFeature.NshmFault(dipFeature));
      }
      return treeBuilder;// .build();
    }

    /* This exists for consistency with legacy codes */
    @Deprecated
    private static final double TO_RAD = 3.14159 / 180.0;

    @Deprecated
    private static double computeWidth(double Δz, double dip) {
      // Inexplicable rounding up of 45° where other dips go down
      RoundingMode rMode = (dip == 45.0) ? RoundingMode.CEILING : RoundingMode.FLOOR;
      return Maths.round(Δz / Math.sin(dip * TO_RAD), 5, rMode);
    }

    @Deprecated
    private void loadSingleRuptureSet(
        Branch<Path> branch,
        SourceTree.Builder treeBuilder,
        ModelData data,
        SourceFeature.NshmFault feature) {

      Map<String, LogicTree<Mfd.Properties>> mfdMap = data.mfdMap().orElseThrow();
      // TODO singleton map is a bad assumtion,
      LogicTree<Mfd.Properties> mfdTree = mfdMap.values().iterator().next();
      data.faultFeatureMap(Map.of(feature.id, feature));

      FaultRuptureSet ruptureSet = FaultRuptureSet.builder()
          .modelData(data)
          .name(feature.name)
          .id(feature.id)
          .sections(List.of(feature.id))
          .mfdTree(mfdTree)
          .build();

      treeBuilder.addLeaf(branch, ruptureSet);
    }

    private FaultRuptureSet createSingleRuptureSet(
        // Branch<Path> branch,
        // SourceTree.Builder treeBuilder,
        ModelData data,
        SourceFeature.NshmFault feature) {

      Map<String, LogicTree<Mfd.Properties>> mfdMap = data.mfdMap().orElseThrow();
      // TODO singleton map is a bad assumtion,
      LogicTree<Mfd.Properties> mfdTree = mfdMap.values().iterator().next();
      data.faultFeatureMap(Map.of(feature.id, feature));

      return FaultRuptureSet.builder()
          .modelData(data)
          .name(feature.name)
          .id(feature.id)
          .sections(List.of(feature.id))
          .mfdTree(mfdTree)
          .build();

      // treeBuilder.addLeaf(branch, ruptureSet);
    }

    @Override
    SourceTree loadSourceTree(
        Path dir,
        ModelData data) {

      System.out.println("    tree: " + root.relativize(dir));
      LogicTree<Path> tree = readSourceTree(dir).orElseThrow();

      /* Set flag to process tree of single geojson sources. */
      boolean singleSourceTree = tree.get(0).value().toString().endsWith(".geojson");

      /* Set flag to process a logic-tree of fault-system solutions. */
      var faultSections = readFaultFeatures(dir);
      faultSections.ifPresent(data::faultFeatureMap);
      boolean faultSystemTree = faultSections.isEmpty() && !singleSourceTree;

      SourceTree.Builder treeBuilder = SourceTree.builder()
          .name(checkNotNull(dir.getFileName()).toString())
          .type(faultSystemTree ? FAULT_SYSTEM : FAULT)
          .gmms(data.gmms())
          .root(tree);

      for (Branch<Path> branch : tree) {
        if (singleSourceTree) {
          processSingleBranch(branch, treeBuilder, data);
        } else if (faultSystemTree) {
          processSystemBranch(branch, treeBuilder, data);
        } else {
          processBranch(branch, treeBuilder, data);
        }
      }
      return treeBuilder.build();
    }

    /* Process fault source tree branches. */
    private void processBranch(
        Branch<Path> branch,
        SourceTree.Builder treeBuilder,
        ModelData dataIn) {

      Path dir = branch.value();
      System.out.println("  branch: " + root.relativize(dir));

      if (checkNotNull(dir.getFileName()).toString().equals("null")) {
        System.out.println("TODO null branch");
        return;
      }

      ModelData data = ModelData.copyOf(dataIn);
      readMfdConfig(dir).ifPresent(data::mfdConfig);

      /*
       * Don't use loadRateTree() here; nested rate trees on branches are more
       * likely numeric recurrence rates than slip rate branches.
       */
      readRateTree(dir).ifPresent(data::rateTree);

      /* Can be source-tree or source-group. */
      Optional<LogicTree<Path>> tree = readSourceTree(dir);
      if (tree.isPresent()) {

        LogicTree<Path> children = tree.orElseThrow();
        treeBuilder.addBranches(branch, children);

        for (Branch<Path> child : children) {
          processBranch(child, treeBuilder, data);
        }

      } else {

        RuptureSet rs = readFaultRuptureSet(dir, data).orElseGet(
            () -> readFaultClusterSet(dir, data).orElseThrow());
        treeBuilder.addLeaf(branch, rs);
      }
    }

    /* Process fault system source tree branches. */
    private void processSystemBranch(
        Branch<Path> branch,
        SourceTree.Builder treeBuilder,
        ModelData data) {

      /*
       * No configuration overrides are currently expected for
       * fault-system-solutions so this is simpler than above.
       */
      Path dir = branch.value();
      System.out.println("  system: " + root.relativize(dir));
      Optional<LogicTree<Path>> tree = readSourceTree(dir);
      if (tree.isPresent()) {
        LogicTree<Path> children = tree.orElseThrow();
        treeBuilder.addBranches(branch, children);
        for (Branch<Path> child : children) {
          processSystemBranch(child, treeBuilder, data);
        }
      } else {
        treeBuilder.addLeaf(
            branch,
            readSystemRuptureSet(dir, data).orElseThrow());
      }
    }

    /*
     * Process branhces consisting of individual geojson files as would be
     * processed using loadSingleSource().
     */
    private void processSingleBranch(
        Branch<Path> branch,
        SourceTree.Builder treeBuilder,
        ModelData data) {

      if (checkNotNull(branch.value().getFileName()).toString().equals("null")) {
        System.out.println("TODO null branch");
        return;
      }

      SourceFeature.NshmFault feature = SourceFeature.newNshmFault(branch.value());
      loadSingleSource(treeBuilder, Optional.of(branch), feature, data);
    }
  }

  static List<SourceTree> interfaceSources(Path dir) {
    GmmSet gmms = readGmms(dir).orElseThrow();
    SourceConfig srcConfig = readInterfaceConfig(dir).orElseThrow();
    ModelData data = ModelData.create(gmms, srcConfig);
    return new Interface(dir).loadSourceDirectory(dir, data);
  }

  private static class Interface extends ModelLoader {

    Interface(Path root) {
      super(root);
      logRoot(root, INTERFACE);
    }

    @Override
    void loadSourceConfig(Path dir, ModelData data) {
      readInterfaceConfig(dir).ifPresent(data::sourceConfig);
    }

    @Override
    void loadRateTree(Path dir, ModelData data) {
      readRateTree(dir).ifPresent(data::rateTree);
    }

    @Override
    SourceTree loadSingleSource(
        Path geojson,
        ModelData data) {

      System.out.println("  source: " + root.relativize(geojson));
      SourceFeature.Interface feature = SourceFeature.newInterfaceSection(geojson);

      // TODO dummy mfds for now TODO TODO TODO
      System.out.println(" -- STOP -- this isnt working correctly");
      Mfd.Properties.Single mfdProps = new Mfd.Properties.Single(7.0);

      LogicTree<Mfd.Properties> mfdTree = LogicTree.singleton(
          Deserialize.MFD_TREE,
          "changeme",
          mfdProps);

      InterfaceRuptureSet ruptureSet = InterfaceRuptureSet.builder()
          .modelData(data)
          .name(feature.name)
          .id(feature.id)
          .sections(List.of(feature.id))
          .mfdTree(mfdTree)
          .build();

      LogicTree<Path> root = LogicTree.singleton(
          Deserialize.SOURCE_TREE,
          feature.name,
          Path.of(feature.name));
      SourceTree tree = SourceTree.builder()
          .name(feature.name)
          .type(INTERFACE)
          .gmms(data.gmms())
          .root(root)
          .addLeaf(root.get(0), ruptureSet)
          .build();

      return tree;
    }

    @Override
    SourceTree loadSourceTree(
        Path dir,
        ModelData data) {

      System.out.println("    tree: " + root.relativize(dir));
      LogicTree<Path> tree = readSourceTree(dir).orElseThrow();
      data.interfaceFeatureMap(readInterfaceFeatures(dir).orElseThrow());

      SourceTree.Builder treeBuilder = SourceTree.builder()
          .name(checkNotNull(dir.getFileName()).toString())
          .type(INTERFACE)
          .gmms(data.gmms())
          .root(tree);

      for (Branch<Path> branch : tree) {
        processBranch(branch, treeBuilder, ModelData.copyOf(data));
      }

      return treeBuilder.build();
    }

    /* Process interface source tree branches. */
    private void processBranch(
        Branch<Path> branch,
        SourceTree.Builder treeBuilder,
        ModelData dataIn) {

      Path dir = branch.value();
      System.out.println("  branch: " + root.relativize(dir));

      if (checkNotNull(dir.getFileName()).toString().equals("null")) {
        System.out.println("TODO null branch");
        return;
      }

      ModelData data = ModelData.copyOf(dataIn);
      readRateTree(dir).ifPresent(data::rateTree);

      /* Can be source-tree or source-group. */
      Optional<LogicTree<Path>> tree = readSourceTree(dir);
      if (tree.isPresent()) {

        LogicTree<Path> children = tree.orElseThrow();
        treeBuilder.addBranches(branch, children);

        for (Branch<Path> child : children) {
          processBranch(child, treeBuilder, data);
        }

      } else {
        treeBuilder.addLeaf(
            branch,
            readInterfaceRuptureSet(dir, data).orElseThrow());
      }
    }
  }

  static List<SourceTree> gridSources(Path dir, GmmSet gmms) {
    SourceConfig srcConfig = readGridConfig(dir).orElseThrow();
    ModelData data = ModelData.create(gmms, srcConfig);
    /* Grid data directory required at grid-sources root. */
    Path gridDataDir = checkGridDataDirectory(dir).orElseThrow();
    data.gridDataDirectory(gridDataDir);
    return new Grid(dir).loadSourceDirectory(dir, data);
  }

  private static class Grid extends ModelLoader {

    Grid(Path root) {
      super(root);
      logRoot(root, GRID);
    }

    @Override
    void loadSourceConfig(Path dir, ModelData data) {
      readGridConfig(dir).ifPresent(data::sourceConfig);
    }

    @Override
    void loadRateTree(Path dir, ModelData data) {
      readGridRateTree(dir).ifPresent(data::gridRateTree);
    }

    @Override
    SourceTree loadSingleSource(
        Path geojson,
        ModelData data) {

      // TODO can this be passed to processBranch as a singleton

      System.out.println("  source: " + root.relativize(geojson));

      Path dir = checkNotNull(geojson.getParent());
      data.gridFeatureMap(ModelFiles.readGridFeatures(dir).orElseThrow());
      LogicTree<Path> rateTree = data.gridRateTree().orElseThrow();

      SourceTree.Builder tree = SourceTree.builder()
          .name(checkNotNull(geojson.getFileName()).toString().split("\\.")[0])
          .type(GRID)
          .gmms(data.gmms())
          .root(rateTree);

      for (Branch<Path> rateBranch : rateTree) {

        /* Set rates file. */
        data.gridRateFile(rateBranch.value());

        /* Create rupture sets, one for each polygon feature. */
        List<GridRuptureSet> ruptureSets = GridLoader.createGridRuptureSets(data);
        checkState(ruptureSets.size() == 1);

        tree.addLeaf(rateBranch, ruptureSets.get(0));
      }
      return tree.build();
    }

    @Override
    SourceTree loadSourceTree(
        Path dir,
        ModelData data) {

      System.out.println("    tree: " + root.relativize(dir));
      LogicTree<Path> tree = readSourceTree(dir).orElseThrow();

      SourceTree.Builder treeBuilder = SourceTree.builder()
          .name(checkNotNull(dir.getFileName()).toString())
          .type(GRID)
          .gmms(data.gmms())
          .root(tree);

      for (Branch<Path> branch : tree) {
        processBranch(branch, treeBuilder, data);
      }

      return treeBuilder.build();
    }

    /* Process grid source tree branches. */
    private void processBranch(
        Branch<Path> branch,
        SourceTree.Builder treeBuilder,
        ModelData dataIn) {

      Path dir = branch.value();
      System.out.println("  branch: " + root.relativize(dir));

      if (checkNotNull(dir.getFileName()).toString().equals("null")) {
        System.out.println("TODO null branch");
        return;
      }

      ModelData data = ModelData.copyOf(dataIn);
      loadRateTree(dir, data);

      Optional<LogicTree<Path>> tree = readSourceTree(dir);
      if (tree.isPresent()) {

        LogicTree<Path> children = tree.orElseThrow();
        treeBuilder.addBranches(branch, children);

        for (Branch<Path> child : children) {
          processBranch(child, treeBuilder, data);
        }

      } else {

        /* Expect one or more geojson polygon files. */
        data.gridFeatureMap(ModelFiles.readGridFeatures(dir).orElseThrow());
        LogicTree<Path> rateTree = data.gridRateTree().orElseThrow();
        treeBuilder.addBranches(branch, rateTree);

        for (Branch<Path> rateBranch : rateTree) {

          /* Set rates file. */
          data.gridRateFile(rateBranch.value());

          /* Create rupture sets, one for each polygon feature. */
          List<GridRuptureSet> ruptureSets = GridLoader.createGridRuptureSets(data);

          /* Attach singleton as leaf. */
          if (ruptureSets.size() == 1) {
            treeBuilder.addLeaf(rateBranch, ruptureSets.get(0));
            continue;
          }

          /* Otherwise create logic group of rupture set pseudo paths. */
          LogicGroup.Builder<Path> grsGroup = LogicGroup.builder();
          for (GridRuptureSet grs : ruptureSets) {
            grsGroup.addBranch(grs.name(), Path.of(grs.name()), 1.0);
          }
          LogicTree<Path> grsTree = grsGroup.build();
          treeBuilder.addBranches(rateBranch, grsTree);

          for (int j = 0; j < grsTree.size(); j++) {
            treeBuilder.addLeaf(grsTree.get(j), ruptureSets.get(j));
          }
        }
      }
    }
  }

  static List<SourceTree> slabSources(Path dir) {
    GmmSet gmms = readGmms(dir).orElseThrow();
    SourceConfig srcConfig = readSlabConfig(dir).orElseThrow();
    ModelData data = ModelData.create(gmms, srcConfig);
    return new Slab(dir).loadSourceDirectory(dir, data);
  }

  private static class Slab extends ModelLoader {

    Slab(Path root) {
      super(root);
      logRoot(root, INTRASLAB);
    }

    @Override
    void loadSourceConfig(Path dir, ModelData data) {
      readSlabConfig(dir).ifPresent(data::sourceConfig);
    }

    @Override
    void loadRateTree(Path dir, ModelData data) {
      readGridRateTree(dir).ifPresent(data::gridRateTree);
    }

    @Override
    SourceTree loadSingleSource(
        Path geojson,
        ModelData data) {

      System.out.println("  source: " + root.relativize(geojson));

      Path dir = checkNotNull(geojson.getParent());
      data.gridFeatureMap(ModelFiles.readGridFeatures(dir).orElseThrow());
      LogicTree<Path> rateTree = data.gridRateTree().orElseThrow();

      SourceTree.Builder tree = SourceTree.builder()
          .name(checkNotNull(geojson.getFileName()).toString().split("\\.")[0])
          .type(INTRASLAB)
          .gmms(data.gmms())
          .root(rateTree);

      for (Branch<Path> rateBranch : rateTree) {

        /* Set rates file; slab needs to be resolved to local dir. */
        data.gridRateFile(dir.resolve(rateBranch.value()));

        /* Create rupture sets, one for each polygon feature. */
        List<SlabRuptureSet> ruptureSets = GridLoader.createSlabRuptureSets(data);
        checkState(ruptureSets.size() == 1);

        tree.addLeaf(rateBranch, ruptureSets.get(0));
      }
      return tree.build();
    }

    @Override
    SourceTree loadSourceTree(
        Path dir,
        ModelData data) {

      System.out.println("    tree: " + root.relativize(dir));
      LogicTree<Path> tree = readSourceTree(dir).orElseThrow();

      SourceTree.Builder treeBuilder = SourceTree.builder()
          .name(checkNotNull(dir.getFileName()).toString())
          .type(INTRASLAB)
          .gmms(data.gmms())
          .root(tree);

      for (Branch<Path> branch : tree) {
        processBranch(branch, treeBuilder, data);
      }

      return treeBuilder.build();
    }

    /* Process intraslab source tree branches. */
    private void processBranch(
        Branch<Path> branch,
        SourceTree.Builder treeBuilder,
        ModelData dataIn) {

      Path dir = branch.value();
      System.out.println("  branch: " + root.relativize(dir));

      if (checkNotNull(dir.getFileName()).toString().equals("null")) {
        System.out.println("TODO null branch");
        return;
      }

      ModelData data = ModelData.copyOf(dataIn);
      loadRateTree(dir, data);

      Optional<LogicTree<Path>> tree = readSourceTree(dir);
      if (tree.isPresent()) {

        LogicTree<Path> children = tree.orElseThrow();
        treeBuilder.addBranches(branch, children);

        for (Branch<Path> child : children) {
          processBranch(child, treeBuilder, data);
        }

      } else {

        /* Expect one or more geojson polygon files. */
        data.gridFeatureMap(ModelFiles.readGridFeatures(dir).orElseThrow());
        LogicTree<Path> rateTree = data.gridRateTree().orElseThrow();
        treeBuilder.addBranches(branch, rateTree);

        for (Branch<Path> rateBranch : rateTree) {

          /* Set rates file; slab needs to be resolved to local dir. */
          data.gridRateFile(dir.resolve(rateBranch.value()));

          /* Create rupture sets, one for each polygon feature. */
          List<SlabRuptureSet> ruptureSets = GridLoader.createSlabRuptureSets(data);

          /* Attach singleton as leaf. */
          if (ruptureSets.size() == 1) {
            treeBuilder.addLeaf(rateBranch, ruptureSets.get(0));
            return;
          }

          /* Otherwise create logic group of rupture set pseudo paths. */
          /* TODO slab doesn't get here. */
          LogicGroup.Builder<Path> grsGroup = LogicGroup.builder();
          for (SlabRuptureSet srs : ruptureSets) {
            grsGroup.addBranch(srs.name(), Path.of(srs.name()), 1.0);
          }
          LogicTree<Path> srsTree = grsGroup.build();
          treeBuilder.addBranches(rateBranch, srsTree);

          for (int j = 0; j < srsTree.size(); j++) {
            treeBuilder.addLeaf(srsTree.get(j), ruptureSets.get(j));
          }
        }
      }
    }
  }

  static List<SourceTree> zoneSources(Path dir, GmmSet gmms) {
    SourceConfig srcConfig = readZoneConfig(dir).orElseThrow();
    ModelData data = ModelData.create(gmms, srcConfig);
    return new Zone(dir).loadSourceDirectory(dir, data);
  }

  private static class Zone extends ModelLoader {

    Zone(Path root) {
      super(root);
      logRoot(root, ZONE);
    }

    @Override
    void loadSourceConfig(Path dir, ModelData data) {
      readZoneConfig(dir).ifPresent(data::sourceConfig);
    }

    @Override
    void loadRateTree(Path dir, ModelData data) {
      readRateTree(dir).ifPresent(data::rateTree);
    }

    @Override
    SourceTree loadSingleSource(
        Path geojson,
        ModelData data) {

      System.out.println("  source: " + root.relativize(geojson));
      SourceFeature.Zone feature = SourceFeature.newSourceZone(geojson);

      Properties props = feature.source.properties();
      String ratesFile = props.getString("rate").orElseThrow();

      ZoneRuptureSet.Builder zrsb = ZoneRuptureSet.builder()
          .feature(feature)
          .modelData(data)
          .rates(geojson.resolveSibling(ratesFile));

      // TODO use deserialize
      JsonElement mfdElement = props.getJsonElement("mfd-tree").orElseThrow();
      if (mfdElement.isJsonPrimitive()) {
        zrsb.mfdTree(mfdElement.getAsString());
      } else {
        zrsb.mfdTree(Deserialize.mfdPropertiesTree(mfdElement));
      }

      ZoneRuptureSet ruptureSet = zrsb.build();

      LogicTree<Path> root = LogicTree.singleton(
          Deserialize.SOURCE_TREE,
          feature.name,
          Path.of(feature.name));

      SourceTree tree = SourceTree.builder()
          .name(feature.name)
          .type(ZONE)
          .gmms(data.gmms())
          .root(root)
          .addLeaf(root.get(0), ruptureSet)
          .build();

      return tree;
    }

    @Override
    SourceTree loadSourceTree(
        Path dir,
        ModelData data) {

      System.out.println("    tree: " + root.relativize(dir));
      LogicTree<Path> tree = readSourceTree(dir).orElseThrow();

      SourceTree.Builder treeBuilder = SourceTree.builder()
          .name(checkNotNull(dir.getFileName()).toString())
          .type(ZONE)
          .gmms(data.gmms())
          .root(tree);

      for (Branch<Path> branch : tree) {
        processBranch(branch, treeBuilder, data);
      }

      return treeBuilder.build();
    }

    /* Process zone (area) source tree branches. */
    private void processBranch(
        Branch<Path> branch,
        SourceTree.Builder treeBuilder,
        ModelData dataIn) {

      Path dir = branch.value();
      System.out.println("  branch: " + root.relativize(dir));

      if (checkNotNull(dir.getFileName()).toString().equals("null")) {
        System.out.println("TODO null branch");
        return;
      }

      ModelData data = ModelData.copyOf(dataIn);
      readRateTree(dir).ifPresent(data::rateTree);

      Optional<LogicTree<Path>> tree = readSourceTree(dir);
      if (tree.isPresent()) {

        LogicTree<Path> children = tree.orElseThrow();
        treeBuilder.addBranches(branch, children);

        for (Branch<Path> child : children) {
          processBranch(child, treeBuilder, data);
        }

      } else {

        Path zonePath;
        try (DirectoryStream<Path> sourceStream = newDirectoryStream(branch.value(), "*.geojson")) {
          zonePath = Iterables.getOnlyElement(sourceStream);
        } catch (IOException ioe) {
          throw new RuntimeException(ioe);
        }

        SourceFeature.Zone zone = SourceFeature.newSourceZone(zonePath);

        Properties props = zone.source.properties();
        String ratesFile = props.getString("rate").orElseThrow();

        ZoneRuptureSet.Builder zrsb = ZoneRuptureSet.builder()
            .feature(zone)
            .modelData(data)
            .rates(zonePath.resolveSibling(ratesFile));

        // TODO use deserialize
        JsonElement mfdElement = props.getJsonElement("mfd-tree").orElseThrow();
        if (mfdElement.isJsonPrimitive()) {
          zrsb.mfdTree(mfdElement.getAsString());
        } else {
          zrsb.mfdTree(Deserialize.mfdPropertiesTree(mfdElement));
        }

        ZoneRuptureSet ruptureSet = zrsb.build();
        treeBuilder.addLeaf(branch, ruptureSet);
      }
    }
  }

}
