package gov.usgs.earthquake.nshmp.model;

import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.base.Preconditions.checkState;

import java.util.List;
import java.util.Optional;

import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.mfd.Mfd;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/*
 * Subduction intraslab rupture set implemention.
 *
 * @author U.S. Geological Survey
 *
 */
class SlabRuptureSet implements RuptureSet {

  private final SourceFeature.Grid feature;
  private final ModelData data;

  private final List<Location> locations;
  private final List<Mfd> mfds;
  private final LogicTree<List<Mfd>> mfdsTree;

  private SlabSourceSet sss;

  SlabRuptureSet(Builder builder) {
    this.feature = builder.feature;
    this.data = builder.data;

    this.locations = builder.locations;
    this.mfds = builder.mfds;
    this.mfdsTree = builder.mfdsTree;
  }

  SlabSourceSet sourceSet(double weight) {
    if (sss == null) {
      sss = createSourceSet(weight);
    }
    return sss;
  }

  private SlabSourceSet createSourceSet(double weight) {
    SourceConfig.Grid config = data.sourceConfig().asGrid();
    GridSourceSet delegate = GridLoader.createGrid(
        config,
        feature.source,
        locations,
        mfds,
        mfdsTree,
        Optional.empty(),
        data.gmms(),
        weight);
    return new SlabSourceSet(delegate);
  }

  @Override
  public String name() {
    return feature.name;
  }

  @Override
  public int id() {
    return feature.id;
  }

  @Override
  public int size() {
    return locations.size();
  }

  @Override
  public SourceType type() {
    return SourceType.INTRASLAB;
  }

  @Override
  public Location location(Location site) {
    // TODO Auto-generated method stub
    return null;
  }

  static Builder builder() {
    return new Builder();
  }

  static final class Builder {

    private boolean built = false;

    private SourceFeature.Grid feature;
    private ModelData data;

    private List<Location> locations;
    private List<Mfd> mfds;
    private LogicTree<List<Mfd>> mfdsTree;

    Builder feature(SourceFeature.Grid feature) {
      this.feature = feature;
      return this;
    }

    Builder modelData(ModelData data) {
      this.data = data;
      return this;
    }

    Builder nodeData(
        List<Location> locations,
        List<Mfd> mfds,
        LogicTree<List<Mfd>> mfdsTree) {

      // TODO check sizes

      this.locations = locations;
      this.mfds = mfds;
      this.mfdsTree = mfdsTree;
      return this;
    }

    private void validateAndInit(String label) {
      checkState(!built, "Single use builder");
      checkNotNull(feature, "%s feature", label);
      checkNotNull(data, "%s model data", label);

      checkNotNull(locations, "%s locations", label);
      checkNotNull(mfds, "%s total MFDs", label);
      checkNotNull(mfdsTree, "%s MFDs logic tree", label);

      built = true;
    }

    /* Create a new rupture set. */
    SlabRuptureSet build() {
      validateAndInit("SlabRuptureSet.Builder");
      return new SlabRuptureSet(this);
    }
  }
}
