package gov.usgs.earthquake.nshmp.model;

import static com.google.common.base.Preconditions.checkArgument;
import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.base.Preconditions.checkState;
import static gov.usgs.earthquake.nshmp.Text.checkName;

import java.nio.file.Path;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import com.google.common.collect.Iterables;
import com.google.common.collect.Lists;
import com.google.common.graph.EndpointPair;
import com.google.common.graph.Graphs;
import com.google.common.graph.MutableNetwork;
import com.google.common.graph.Network;
import com.google.common.graph.NetworkBuilder;
import com.google.common.graph.Traverser;

import gov.usgs.earthquake.nshmp.Maths;
import gov.usgs.earthquake.nshmp.tree.Branch;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/**
 * TODO update; Top level wrapper class used for any source logic trees. For
 * example, UCERF3 alternate fault models, New Madrid, Wasatch, or any zone or
 * grid models?
 *
 * @author U.S. Geological Survey
 */
public class SourceTree {

  /*
   * Using optionals for now, but may want base class that handles all the
   * branches with subclasses for different source model types: sections vs
   * polygons vs grids
   *
   * Developer notes:
   *
   * Root --> Node --> Node --> Leaf
   *
   * We use Guava's graph library to track source model logic trees. The
   * heirarchy of logic trees in a source model is mapped into a Network where
   * nodes are connected with unique edges that are the branches of the
   * component logic trees.
   *
   * We tried using a regular Graph with Branches representing the nodes, but it
   * was difficult to extract the node paths between root and Leaf endpoints.
   *
   * I don't hink we need to really deal with the Node class, it's just
   * connectiuve tissue and supports rooot/trunk and leaf
   *
   * With regards to sampling, we may want to consider holding a reference to
   * the relevant logic tree at each node.
   *
   * The internal graph (Network) is a directed tree; each node has only one
   * incident edge/node.
   *
   * Leaf weights are rounded to 6 decimal places.
   */

  private final String name;
  private final SourceType type;
  private final GmmSet gmms;

  // TODO LogicTree<Leaf> ???
  // SHould leaf also store it's path??

  // TODO enforce collection immutability; currently some branches
  // are null preventing this; perhaps have a EmptyNode with no path
  // TODO may not need to keep Network (should also be immutable if we do)
  private final Network<Node, Branch<Path>> tree;
  private final Map<Leaf, List<Branch<Path>>> leafBranches;
  private final Map<Leaf, Double> leafWeights;

  /** The name of this tree. */
  public String name() {
    return name;
  }

  /** The type of sources in this tree. */
  public SourceType type() {
    // TODO what should this return for mixed fault/cluster ??
    return type;
  }

  /** The ground motion models to use with this tree. */
  public GmmSet gmms() {
    return gmms;
  }

  public Map<Leaf, List<Branch<Path>>> branchMap() {
    return leafBranches;
  }

  public Map<Leaf, Double> branchWeights() {
    return leafWeights;
  }

  SourceTree(Builder builder) {
    this.name = builder.name;
    this.type = builder.type;
    this.gmms = builder.gmms;
    this.tree = builder.tree;
    this.leafBranches = builder.leafBranches;
    this.leafWeights = builder.leafWeights;

    // TODO clean
    // for (Leaf leaf : leafBranches.keySet()) {
    // System.out.println(leafBranches.get(leaf) + " " + leafWeights.get(leaf));
    // }
  }

  static Builder builder() {
    return new Builder();
  }

  private static final Root ROOT = new Root();

  static class Builder {

    private boolean built = false;

    private int index = 1;

    /* Required fields. */
    private String name;
    private SourceType type;

    // TODO currently visible for building SourceSets
    // make private when SourceTree replaces SourceSet
    GmmSet gmms;

    private MutableNetwork<Node, Branch<Path>> tree;

    /* Created on build. */
    private Map<Leaf, List<Branch<Path>>> leafBranches;
    private Map<Leaf, Double> leafWeights;

    private List<Leaf> leaves = new ArrayList<>();

    Builder() {} // TODO make private

    /* Set the name of the tree. */
    Builder name(String name) {
      checkState(this.name == null, "Tree name already set");
      this.name = checkName(name, "Tree name");
      return this;
    }

    /* Set the type of source in the tree. */
    Builder type(SourceType type) {
      checkState(this.type == null, "Tree source type already set");
      this.type = checkNotNull(type);
      return this;
    }

    /* Set the GMMs to be used with the sources in the tree. */
    Builder gmms(GmmSet gmms) {
      checkState(this.gmms == null, "Tree GMMs already set");
      this.gmms = checkNotNull(gmms);
      return this;
    }

    /* Set the logic tree that stems from the root node. */
    Builder root(LogicTree<Path> root) {
      checkState(this.tree == null, "Tree root node already set");
      this.tree = NetworkBuilder.directed()
          .allowsSelfLoops(false)
          .allowsParallelEdges(false)
          .build();
      for (Branch<Path> branch : root) {
        tree.addEdge(ROOT, new Node(index++), branch);
      }
      return this;
    }

    /* Attach the branches of a logic tree to the supplied branch. */
    Builder addBranches(Branch<Path> parent, LogicTree<Path> children) {
      checkState(this.tree != null, "Tree root node is not set");
      checkArgument(tree.edges().contains(parent), "Missing parent branch");
      EndpointPair<Node> endpoints = tree.incidentNodes(parent);
      // parent target is new source for each branch being added
      Node source = endpoints.target();
      for (Branch<Path> child : children) {
        tree.addEdge(source, new Node(index++), child);
      }
      return this;
    }

    /*
     * Add a leaf to the source tree. This method replaces the target Node
     * associated with the supplied branch with a terminal Leaf node.
     */
    Builder addLeaf(Branch<Path> parent, RuptureSet ruptureSet) {
      checkState(this.tree != null, "Tree root node is not set");
      EndpointPair<Node> endpoints = tree.incidentNodes(parent);
      Node target = endpoints.target();
      Node source = endpoints.source();
      tree.removeNode(target);
      Leaf leaf = new Leaf(target.index, ruptureSet); // transfer assigned index
      tree.addEdge(source, leaf, parent);
      leaves.add(leaf);
      return this;
    }

    /* Create a new source tree from this builder. */
    SourceTree build() {
      checkState(!built, "Single use builder");
      built = true;
      checkNotNull(name);
      checkNotNull(type);
      checkNotNull(gmms);
      checkState(tree.nodes().size() > 0, "Empty tree");
      List<Leaf> leafList = List.copyOf(leaves);
      checkState(leafList.size() > 0, "Leafless tree");
      leafBranches = branchLists(tree, leafList);
      leafWeights = leafWeights(leafBranches);
      return new SourceTree(this);
    }
  }

  /*
   * On build: Generate lists of logic tree branches (branch paths) from the
   * root of a tree to each of the supplied leaf nodes. Because the tree is a
   * directed graph, traversing a transposed view starting with each leaf yields
   * the required node path.
   */
  private static Map<Leaf, List<Branch<Path>>> branchLists(
      Network<Node, Branch<Path>> tree,
      List<Leaf> leaves) {

    Traverser<Node> traverser = Traverser.forTree(Graphs.transpose(tree));
    Map<Leaf, List<Branch<Path>>> branchListMap = new HashMap<>();
    for (Leaf leaf : leaves) {
      List<Node> nodeList = Lists.newArrayList(traverser.depthFirstPostOrder(leaf));
      checkState(nodeList.size() > 1); // 2 nodes minimum [root -> leaf]
      List<Branch<Path>> branchList = new ArrayList<>();
      Node source = nodeList.get(0);
      for (Node target : Iterables.skip(nodeList, 1)) {
        branchList.add(tree.edgeConnecting(source, target).orElseThrow());
        source = target;
      }
      branchListMap.put(leaf, List.copyOf(branchList));
    }
    return Map.copyOf(branchListMap);
  }

  /*
   * On build: Create map of leaves and their weights. Note that the values of
   * the returned map will not sum to one when the source tree contains one or
   * more LogicGroups.
   */
  private static Map<Leaf, Double> leafWeights(Map<Leaf, List<Branch<Path>>> branchLists) {
    return branchLists.entrySet().stream()
        .collect(Collectors.toUnmodifiableMap(
            e -> e.getKey(),
            e -> leafWeight(e.getValue())));
  }

  /* Compute cumulative weight of a source branch from root to leaf. */
  private static double leafWeight(List<Branch<Path>> branchList) {
    return Maths.round(branchList.stream()
        .mapToDouble(Branch::weight)
        .reduce(1, (a, b) -> a * b),
        6);
  }

  static class Node {

    int index;

    Node(int index) {
      this.index = index;
    }

    @Override
    public String toString() {
      return "node:" + index;
    }
  }

  static class Root extends Node {

    Root() {
      super(0);
    }

    @Override
    public String toString() {
      return "root:" + index;
    }
  }

  static class Leaf extends Node {

    // TODO can a Leaf carry the final weight to applied to the RS

    RuptureSet ruptureSet;

    Leaf(int index, RuptureSet ruptureSet) {
      super(index);
      this.ruptureSet = ruptureSet;
    }

    @Override
    public String toString() {
      return "leaf:" + index;
    }

    RuptureSet ruptureSet() {
      return ruptureSet;
    }
  }
}
