package gov.usgs.earthquake.nshmp.model;

import static com.google.common.base.Preconditions.checkArgument;
import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.base.Preconditions.checkState;
import static gov.usgs.earthquake.nshmp.model.SourceFeature.Key.DEPTH;
import static gov.usgs.earthquake.nshmp.model.SourceFeature.Key.DIP;
import static gov.usgs.earthquake.nshmp.model.SourceFeature.Key.INDICES;
import static gov.usgs.earthquake.nshmp.model.SourceFeature.Key.M;
import static gov.usgs.earthquake.nshmp.model.SourceFeature.Key.RAKE;
import static gov.usgs.earthquake.nshmp.model.SourceFeature.Key.RATE;
import static gov.usgs.earthquake.nshmp.model.SourceFeature.Key.WIDTH;
import static gov.usgs.earthquake.nshmp.model.SourceType.FAULT_SYSTEM;
import static java.util.stream.Collectors.toUnmodifiableList;

import java.nio.file.Path;
import java.util.List;
import java.util.Map;
import java.util.stream.IntStream;
import java.util.stream.Stream;

import gov.usgs.earthquake.nshmp.data.DelimitedData;
import gov.usgs.earthquake.nshmp.data.DelimitedData.Record;
import gov.usgs.earthquake.nshmp.fault.surface.DefaultGriddedSurface;
import gov.usgs.earthquake.nshmp.fault.surface.GriddedSurface;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.model.SourceFeature.SystemSection;

/**
 * Crustal fault-system rupture set.
 *
 * @author U.S. Geological Survey
 */
public class SystemRuptureSet implements RuptureSet {

  /*
   * TODO Work is required to improve indexing of fault systems. Currently The
   * two UCERF3 fault model subsections consist of overlapping sets of IDs most
   * of which are common to both models. There are also many shared ruptures
   * across branches of the UCERF3 logic tree, but it is not clear if averaged
   * properties are consistent.
   *
   * For now, the fault model subsection IDs have been bumped up to 31**** and
   * 32**** until we can work on reducing redundancy. There should then be a
   * mapping in rupture-set between index and relevant fault section. This will
   * better support more logic tree branches beyond the two fault models.
   */

  private final ModelData data;
  private final String name;
  private final Integer id;

  private final Map<Integer, SystemSection> sectionMap;
  private final Path ruptures;

  private SystemSourceSet sourceSet;

  SystemRuptureSet(Builder builder) {
    this.data = builder.data;
    this.name = builder.name;
    this.id = builder.id;
    this.sectionMap = builder.sectionMap;
    this.ruptures = builder.ruptures;
  }

  SystemSourceSet createSourceSet(double branchWeight) {
    if (sourceSet == null) {

      SystemSourceSet.Builder builder = new SystemSourceSet.Builder();

      builder.name(name)
          .id(id)
          .weight(branchWeight)
          .gmms(data.gmms());

      List<GriddedSurface> sections = sectionMap.keySet().stream()
          .sorted()
          .map(sectionMap::get)
          .map(SystemRuptureSet::featureToSurface)
          .collect(toUnmodifiableList());
      builder.sections(sections);

      List<String> names = sectionMap.keySet().stream()
          .sorted()
          .map(sectionMap::get)
          .map(feature -> feature.name)
          .collect(toUnmodifiableList());
      builder.sectionNames(names);

      DelimitedData.comma(ruptures).records()
          .forEach(record -> addRupture(builder, record));

      sourceSet = builder.build();

    }
    return sourceSet;
  }

  private static void addRupture(
      SystemSourceSet.Builder builder,
      Record record) {

    builder.mag(record.getDouble(M))
        .rate(record.getDouble(RATE))
        .indices(ruptureIndexStringToIndices(record.get(INDICES)))
        .depth(record.getDouble(DEPTH))
        .dip(record.getDouble(DIP))
        .rake(record.getDouble(RAKE))
        .width(record.getDouble(WIDTH));
  }

  // TODO move to Indexing?
  private static int[] ruptureIndexStringToIndices(String s) {
    return Stream.of(s.split("-"))
        .flatMapToInt(SystemRuptureSet::toIndices)
        .toArray();
  }

  private static IntStream toIndices(String s) {
    return s.contains(":") ? rangeToIndices(s) : IntStream.of(Integer.valueOf(s));
  }

  /*
   * Must be string of the form #:#
   *
   * NOTE: the string provided to ruptureIndexStringToIndices replicates the
   * order of the declared subsections in the original UCERF3 solutions. This
   * stitching order is ont relevant for how nshmp-lib processes hazard and in
   * the method below, the order of indices is reversed to always be ascending.
   */
  private static IntStream rangeToIndices(String s) {
    String[] minmax = s.split(":");
    int min = Integer.valueOf(minmax[0]);
    int max = Integer.valueOf(minmax[1]);
    if (min > max) {
      int newMax = min;
      min = max;
      max = newMax;
    }
    return IntStream.rangeClosed(min, max);
  }

  private static GriddedSurface featureToSurface(SourceFeature.SystemSection section) {
    return DefaultGriddedSurface.builder()
        .depth(section.upperDepth)
        .lowerDepth(section.lowerDepth)
        .aseis(section.aseismicity)
        .dip(section.dip)
        .dipDir(section.dipDirection)
        .trace(section.trace)
        .build();
  }

  @Override
  public String name() {
    return name;
  }

  @Override
  public int id() {
    return id;
  }

  @Override
  public int size() {
    return sourceSet.size();
  }

  @Override
  public SourceType type() {
    return FAULT_SYSTEM;
  }

  /**
   * The closest point on the fault trace, relative to the supplied site
   * {@code Location}.
   */
  @Override
  public Location location(Location site) {
    return null; // TODO Locations.closestPoint(site, feature.trace);
  }

  @Override
  public String toString() {
    return getClass().getSimpleName() + " " + Map.of(
        "name", name(),
        "id", id());// , TODO update
    // "feature", feature.source.toJson());
  }

  static Builder builder() {
    return new Builder();
  }

  // TODO temporary name generating method, called externally
  // makes assumptions about ucerf3 and two fault models
  // TODO spotless path null checks are reallyannoying
  static String ruptureSetName(Path branchPath) {
    String parent = checkNotNull(checkNotNull(branchPath.getParent()).getFileName()).toString();
    String branch = checkNotNull(branchPath.getFileName()).toString();
    return parent + "-" + branch;
  }

  /* Single use builder */
  static class Builder {

    private boolean built = false;

    private ModelData data; // TODO needed?

    private String name;
    private Integer id;
    private Map<Integer, SystemSection> sectionMap;
    private Path ruptures;

    Builder name(String name) {
      this.name = name;
      return this;
    }

    Builder id(int id) {
      this.id = id;
      return this;
    }

    Builder sections(Map<Integer, SystemSection> sectionMap) {
      checkArgument(checkNotNull(sectionMap).size() >= 1);
      this.sectionMap = sectionMap;
      return this;
    }

    Builder ruptures(Path ruptures) {
      this.ruptures = checkNotNull(ruptures);
      return this;
    }

    /* Set MFD helper data. */
    Builder modelData(ModelData data) {
      this.data = checkNotNull(data);
      return this;
    }

    SystemRuptureSet build() {
      validateAndInit("SystemRuptureSet.Builder");
      return new SystemRuptureSet(this);
    }

    private void validateAndInit(String label) {
      checkState(!built, "Single use builder");
      checkNotNull(name, "%s name", label);
      checkNotNull(id, "%s id", label);
      checkNotNull(data, "%s model data", label);
      checkNotNull(sectionMap, "%s subsection map", label);
      checkNotNull(ruptures, "%s rupture file", label);
      built = true;
    }
  }

}
