package gov.usgs.earthquake.nshmp.model;

import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.base.Preconditions.checkState;
import static gov.usgs.earthquake.nshmp.model.Deserialize.MFD_TREE;
import static gov.usgs.earthquake.nshmp.model.Deserialize.RATE_TREE;
import static java.util.stream.Collectors.toUnmodifiableList;

import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Optional;

import gov.usgs.earthquake.nshmp.data.DoubleData;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.geo.Locations;
import gov.usgs.earthquake.nshmp.mfd.Mfd;
import gov.usgs.earthquake.nshmp.tree.Branch;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/*
 * Areal source zone rupture set implemention.
 *
 * @author U.S. Geological Survey
 */
class ZoneRuptureSet implements RuptureSet {

  private final SourceFeature.Zone feature;
  private final ModelData data;

  private final List<Location> locations;
  private final List<Mfd> mfds;
  private final LogicTree<List<Mfd>> mfdsTree;

  private GridSourceSet gss;

  /*
   * TODO Move to setting rate-tree; if single value "rate" present, create
   * singleton rate-tree.
   *
   * Rates in rate tree will need to be distributed to grid rates that scale
   * with latitude.
   */

  ZoneRuptureSet(Builder builder) {
    this.feature = builder.feature;
    this.data = builder.data;
    this.locations = builder.locations;
    this.mfdsTree = builder.mfdsTree;
    this.mfds = builder.mfds;
  }

  // TODO declare this in RuptureSet? Source T issue
  GridSourceSet sourceSet(double weight) {
    if (gss == null) {
      gss = createSourceSet(weight);
    }
    return gss;
  }

  private GridSourceSet createSourceSet(double weight) {
    SourceConfig.Grid config = data.sourceConfig().asGrid();

    GridSourceSet grid = GridLoader.createGrid(
        config,
        feature.source,
        locations,
        mfds,
        mfdsTree,
        Optional.empty(),
        data.gmms(),
        weight);

    return grid;
  }

  @Override
  public String name() {
    return feature.name;
  }

  @Override
  public int id() {
    return feature.id;
  }

  @Override
  public int size() {
    return locations.size();
  }

  @Override
  public SourceType type() {
    return SourceType.ZONE;
  }

  @Override
  public Location location(Location site) {
    // TODO is this called a lot and should be precomputed??
    return Locations.centroid(locations);
  }

  static Builder builder() {
    return new Builder();
  }

  static final class Builder {

    private boolean built = false;

    private SourceFeature.Zone feature;
    private ModelData data;
    private Path ratesPath;
    private LogicTree<Mfd.Properties> mfdPropertiesTree;

    private Optional<String> mfdTreeKey = Optional.empty();

    private List<Location> locations;
    private LogicTree<List<Mfd>> mfdsTree;
    private List<Mfd> mfds;

    private LogicTree<double[]> ratesTree;

    Builder feature(SourceFeature.Zone feature) {
      this.feature = feature;
      return this;
    }

    Builder modelData(ModelData data) {
      this.data = data;
      return this;
    }

    // absolute path to rates
    Builder rates(Path ratesPath) {
      this.ratesPath = ratesPath;
      return this;
    }

    /* Set the MFD tree using a mfdMap key. */
    Builder mfdTree(String key) {
      this.mfdTreeKey = Optional.of(key);
      return this;
    }

    /* Set the MFD tree. */
    Builder mfdTree(LogicTree<Mfd.Properties> mfdTree) {
      this.mfdPropertiesTree = mfdTree;
      return this;
    }

    private void validateAndInit(String label) {
      checkState(!built, "Single use builder");
      checkNotNull(data, "%s model data", label);
      checkNotNull(feature, "%s feature", label);
      checkNotNull(ratesPath, "%s rates file path", label);
      if (mfdTreeKey.isPresent()) {
        String key = mfdTreeKey.orElseThrow();
        mfdPropertiesTree = data.mfdMap().orElseThrow().get(key);
      }
      checkNotNull(mfdPropertiesTree, "%s MFD tree not set", label);
      initLocationAndRateLists();
      mfds = createTotalMfds();
      mfdsTree = createSingleMfdsTree();
    }

    private void initLocationAndRateLists() {
      /*
       * TODO this will change to use JA's rate distribution code
       */
      List<Location> locations = new ArrayList<>();
      List<Double> rates = new ArrayList<>();
      GridLoader.readZoneRateFile(ratesPath, locations, rates);
      this.locations = List.copyOf(locations);
      double[] rateArray = rates.stream()
          .mapToDouble(Double::doubleValue)
          .toArray();
      this.ratesTree = LogicTree.singleton(RATE_TREE, "rate", rateArray);
    }

    /* Create a new rupture set. */
    ZoneRuptureSet build() {
      validateAndInit("ZoneRuptureSet.Builder");
      return new ZoneRuptureSet(this);
    }

    // TODO to builder
    private List<Mfd> createTotalMfds() {
      // TODO hack to get WUS GR zones to work
      double[] rates = reduceRatesTree();
      Mfd.Properties props0 = mfdPropertiesTree.get(0).value();
      if (props0.type().equals(Mfd.Type.SINGLE)) {
        Mfd model = MfdTrees.reduceSingleMfdTree(mfdPropertiesTree);
        return Arrays.stream(rates)
            .mapToObj(r -> Mfd.Builder.from(model).scale(r).build())
            .collect(toUnmodifiableList());
      }
      // else GR
      return Arrays.stream(rates)
          .mapToObj(
              r -> Mfd.Properties.GutenbergRichter.copyOfWithRate(props0.getAsGr(), r)
                  .toBuilder()
                  .build())
          .collect(toUnmodifiableList());
    }

    private double[] reduceRatesTree() {
      double[] reduced = new double[locations.size()];
      double[] scaled = new double[locations.size()];
      for (Branch<double[]> branch : ratesTree) {
        scaled = Arrays.copyOf(branch.value(), branch.value().length);
        DoubleData.multiply(branch.weight(), scaled);
        DoubleData.add(reduced, scaled);
      }
      return reduced;
    }

    // SINGLE only
    private LogicTree<List<Mfd>> createSingleMfdsTree() {
      Mfd.Properties props0 = mfdPropertiesTree.get(0).value();
      // TODO hack short circuit for WUS zones
      if (props0.type().equals(Mfd.Type.GR)) {
        return null;
      }
      LogicTree.Builder<List<Mfd>> mfdsTree = LogicTree.builder(MFD_TREE);
      for (Branch<Mfd.Properties> m : mfdPropertiesTree) {
        Mfd.Builder builder = m.value().getAsSingle().toBuilder();
        for (Branch<double[]> r : ratesTree) {
          List<Mfd> mfds = Arrays.stream(r.value())
              .mapToObj(builder::scaleToCumulativeRate)
              .map(Mfd.Builder::build)
              .collect(toUnmodifiableList());
          mfdsTree.addBranch(
              m.id() + "-" + r.id(),
              mfds,
              m.weight() * r.weight());
        }
      }
      return mfdsTree.build();
    }

  }
}
