package gov.usgs.earthquake.nshmp;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertSame;
import static org.junit.jupiter.api.Assertions.assertThrows;

import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.ValueSource;

import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.geo.LocationList;
import gov.usgs.earthquake.nshmp.geo.Locations;

final class FaultsTests {

  @ParameterizedTest
  @ValueSource(doubles = { 10.0, 45.0, 90.0 })
  final void testCheckDip_inRange(double v) {
    assertEquals(v, Faults.checkDip(v), 0.0);
  }

  @ParameterizedTest
  @ValueSource(doubles = { -30.0, 9.9, 90.1, 100.0 })
  final void testCheckDip_outOfRange(double v) {
    assertThrows(IllegalArgumentException.class, () -> {
      Faults.checkDip(v);
    });
  }

  @ParameterizedTest
  @ValueSource(doubles = { -180.0, 0.0, 180.0 })
  final void testCheckRake_inRange(double v) {
    assertEquals(v, Faults.checkRake(v), 0.0);
  }

  @ParameterizedTest
  @ValueSource(doubles = { -360.0, -180.1, 180.1, 1000.0 })
  final void testCheckRake_outOfRange(double v) {
    assertThrows(IllegalArgumentException.class, () -> {
      Faults.checkRake(v);
    });
  }

  @ParameterizedTest
  @ValueSource(doubles = { 1e-6, 4.0, 60.0 })
  final void testCheckSlipRate_inRange(double v) {
    assertEquals(v, Faults.checkSlipRate(v), 0.0);
  }

  @ParameterizedTest
  @ValueSource(doubles = { -1.0, 0.0, 60.1, 100.0 })
  final void testCheckSlipRate_outOfRange(double v) {
    assertThrows(IllegalArgumentException.class, () -> {
      Faults.checkSlipRate(v);
    });
  }

  @ParameterizedTest
  @ValueSource(doubles = { 0.0, 60.0, 359.9 })
  final void testCheckStrike_inRange(double v) {
    assertEquals(v, Faults.checkStrike(v), 0.0);
  }

  @ParameterizedTest
  @ValueSource(doubles = { -1.0, -0.1, 360.0, 1000.0 })
  final void testCheckStrike_outOfRange(double v) {
    assertThrows(IllegalArgumentException.class, () -> {
      Faults.checkStrike(v);
    });
  }

  @Test
  final void testCheckTrace() {
    LocationList expect = LocationList.of(
        Location.create(0.0, 0.0),
        Location.create(1.0, 1.0));
    assertSame(expect, Faults.checkTrace(expect));
  }

  @Test
  final void testCheckTrace_invalid() {
    assertThrows(IllegalArgumentException.class, () -> {
      Faults.checkTrace(LocationList.of(Location.create(0.0, 0.0)));
    });
  }

  @Test
  final void testHypoDepth() {
    double hypoDepthExpected = 6.0;
    assertEquals(hypoDepthExpected, Faults.hypocentralDepth(90.0, 10.0, 1.0), 0.0);
    hypoDepthExpected = 1.0 + Math.sin(60.0 * Maths.TO_RADIANS) * 10.0 / 2.0;
    assertEquals(hypoDepthExpected, Faults.hypocentralDepth(60.0, 10.0, 1.0), 0.0);
    // 3-4-5 triangle
    hypoDepthExpected = 4.0 + 1.0;
    assertEquals(hypoDepthExpected,
        Faults.hypocentralDepth(Math.asin(4.0 / 5.0) * Maths.TO_DEGREES, 10.0, 1.0), 1e-8);
  }

  /* radian degrees conversion tolerance */
  private static final double RAD_CONV_TOL = 1e-15;

  @Test
  final void testDipDirection() {
    Location p1 = Location.create(0.0, 0.0);
    Location p2 = Location.create(0.0, 1.0);
    Location p3 = Location.create(1.0, 1.0);
    LocationList locs = LocationList.of(p1, p2, p3);

    double strikeExpect = Locations.azimuth(p1, p3);
    double strikeRadExpect = Locations.azimuthRad(p1, p3);

    assertEquals(strikeExpect, Faults.strike(locs), 0.0);
    assertEquals(strikeExpect, Faults.strike(p1, p3), 0.0);
    assertEquals(strikeRadExpect, Faults.strikeRad(locs), 0.0);
    assertEquals(strikeRadExpect, Faults.strikeRad(p1, p3), 0.0);

    double dipDirExpect = 70.0;
    double dipDirRadExpect = dipDirExpect * Maths.TO_RADIANS;

    assertEquals(dipDirExpect, Faults.dipDirection(340.0), 0.0);
    assertEquals(
        dipDirRadExpect,
        Faults.dipDirectionRad(340.0 * Maths.TO_RADIANS), RAD_CONV_TOL);

    dipDirExpect = strikeExpect + 90.0;
    dipDirRadExpect = strikeRadExpect + Maths.PI_BY_2;

    assertEquals(dipDirExpect, Faults.dipDirection(locs), 0.0);
    assertEquals(dipDirExpect, Faults.dipDirection(p1, p3), 0.0);
    assertEquals(dipDirRadExpect, Faults.dipDirectionRad(locs), 0.0);
    assertEquals(dipDirRadExpect, Faults.dipDirectionRad(p1, p3), 0.0);
  }

}
