package gov.usgs.earthquake.nshmp.data;

import static org.junit.jupiter.api.Assertions.assertArrayEquals;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;

import java.util.Arrays;
import java.util.stream.IntStream;

import org.junit.jupiter.api.Test;

import com.google.common.primitives.Doubles;

import gov.usgs.earthquake.nshmp.data.IntervalTable.Builder;

class IntervalTableTests {

  static final double SCALE = 2.0;
  static final double Δ = 1.0;
  static final double ROW_MIN = 2.0;
  static final double ROW_MAX = 6.0;
  static final double COL_MIN = ROW_MIN + SCALE;
  static final double COL_MAX = ROW_MAX + SCALE;

  static final double[] ROWS = IntervalData.keys(ROW_MIN, ROW_MAX, Δ);
  static final double[] COLS = IntervalData.keys(COL_MIN, COL_MAX, Δ);
  static final XySequence XY = XySequence.create(ROWS, COLS);

  static final IntervalTable TABLE;

  static {
    Builder builder = new Builder()
        .rows(ROW_MIN, ROW_MAX, Δ)
        .columns(COL_MIN, COL_MAX, Δ);

    Arrays.stream(ROWS).forEach(row -> builder.add(row, COLS));
    TABLE = builder.build();
  }

  @Test
  final void getTests() {
    IntStream.range(0, ROWS.length).forEach(i -> {
      double row = ROWS[i];
      IntStream.range(0, COLS.length).forEach(j -> {
        double col = COLS[j];
        assertEquals(col, TABLE.get(row, col), 0);
        assertEquals(col, TABLE.get(i, j), 0);
      });
    });

    assertThrows(IndexOutOfBoundsException.class, () -> {
      TABLE.get(1.0, 1.0);
    });
  }

  @Test
  final void rowTests() {
    assertEquals(ROW_MIN, TABLE.rowMin(), 0);
    assertEquals(ROW_MAX, TABLE.rowMax(), 0);
    assertEquals(Δ, TABLE.rowΔ(), 0);
    assertArrayEquals(ROWS, Doubles.toArray(TABLE.rows()), 0);

    IntStream.range(0, ROWS.length).forEach(i -> {
      double row = ROWS[i];

      assertArrayEquals(COLS, TABLE.row(row).xValues().toArray(), 0);
      assertArrayEquals(COLS, TABLE.row(i).xValues().toArray(), 0);

      assertArrayEquals(COLS, TABLE.row(row).yValues().toArray(), 0);
      assertArrayEquals(COLS, TABLE.row(i).yValues().toArray(), 0);
    });
  }

  @Test
  final void columnTests() {
    assertEquals(COL_MIN, TABLE.columnMin(), 0);
    assertEquals(COL_MAX, TABLE.columnMax(), 0);
    assertEquals(Δ, TABLE.columnΔ(), 0);
    assertArrayEquals(COLS, Doubles.toArray(TABLE.columns()), 0);
  }

  @Test
  final void collapseTests() {
    double[] expected = Arrays.stream(COLS).map(col -> Arrays.stream(COLS).sum()).toArray();
    IntervalArray array = TABLE.collapse();
    XySequence values = array.values();

    assertArrayEquals(ROWS, values.xValues().toArray(), 0);
    assertArrayEquals(expected, values.yValues().toArray(), 0);
  }

  @Test
  final void indexTests() {
    double row = ROWS[0];
    double[] valuesA = { 0, 1.0, 2.0, 3.0 };
    double[] valuesB = { 3.0, 2.0, 1.0, 0.0 };
    double[] valuesC = { 1.0, 3.0, 0.0, 2.0 };

    int[] expectedMinA = { 0, 0 };
    int[] expectedMaxA = { 0, 3 };;
    int[] expectedMinB = { 0, 3 };
    int[] expectedMaxB = { 0, 0 };
    int[] expectedMinC = { 0, 2 };
    int[] expectedMaxC = { 0, 1 };

    IntervalTable actualA = Builder.fromModel(TABLE).add(row, valuesA).build();
    IntervalTable actualB = Builder.fromModel(TABLE).add(row, valuesB).build();
    IntervalTable actualC = Builder.fromModel(TABLE).add(row, valuesC).build();

    assertArrayEquals(expectedMinA, actualA.minIndex());
    assertArrayEquals(expectedMaxA, actualA.maxIndex());

    assertArrayEquals(expectedMinB, actualB.minIndex());
    assertArrayEquals(expectedMaxB, actualB.maxIndex());

    assertArrayEquals(expectedMinC, actualC.minIndex());
    assertArrayEquals(expectedMaxC, actualC.maxIndex());
  }

  @Test
  final void toStringTests() {
    String expected = TABLE.toString();
    String actual = Builder.copyOf(TABLE).build().toString();

    assertEquals(expected, actual);
    assertTrue(expected.equals(actual));
  }

  @Test
  final void builderCopyOfTests() {
    IntervalTable actual = Builder.copyOf(TABLE).build();

    assertArrayEquals(TABLE.minIndex(), actual.minIndex());
    assertArrayEquals(TABLE.maxIndex(), actual.maxIndex());

    assertEquals(TABLE.columnMin(), actual.columnMin(), 0);
    assertEquals(TABLE.columnMax(), actual.columnMax(), 0);
    assertEquals(TABLE.columns(), actual.columns());
    assertTrue(TABLE.columns().equals(actual.columns()));
    assertEquals(TABLE.columnΔ(), actual.columnΔ(), 0);

    assertEquals(TABLE.rowMin(), actual.rowMin(), 0);
    assertEquals(TABLE.rowMax(), actual.rowMax(), 0);
    assertEquals(TABLE.rows(), actual.rows());
    assertTrue(TABLE.rows().equals(actual.rows()));
    assertEquals(TABLE.rowΔ(), actual.rowΔ(), 0);

    assertEquals(TABLE.collapse().values(), actual.collapse().values());
  }

  @Test
  final void builderFromModelTests() {
    IntervalTable actual = Builder.fromModel(TABLE).build();

    assertEquals(TABLE.columnMax(), actual.columnMax(), 0);
    assertEquals(TABLE.columnMin(), actual.columnMin(), 0);
    assertEquals(TABLE.columns(), actual.columns());
    assertEquals(TABLE.columnΔ(), actual.columnΔ(), 0);

    assertEquals(TABLE.rowMin(), actual.rowMin(), 0);
    assertEquals(TABLE.rowMax(), actual.rowMax(), 0);
    assertEquals(TABLE.rows(), actual.rows());
    assertEquals(TABLE.rowΔ(), actual.rowΔ(), 0);
  }

  @Test
  final void builderIndexTests() {
    int expected = 3;
    int actualRow = Builder.fromModel(TABLE).rowIndex(ROWS[expected]);
    int actualCol = Builder.fromModel(TABLE).columnIndex(COLS[expected]);

    assertEquals(expected, actualRow);
    assertEquals(expected, actualCol);
  }

  @Test
  final void builderAddTests() {
    int index = 0;
    double[] expected = Arrays.stream(COLS).map(x -> x + x).toArray();

    // add(double row, double[] values)
    IntervalTable actualA = Builder.copyOf(TABLE)
        .add(ROWS[index], COLS)
        .build();

    assertArrayEquals(expected, actualA.row(index).yValues().toArray(), 0);

    // add(double row, XySequence values)
    IntervalTable actualB = Builder.copyOf(TABLE)
        .add(ROWS[index], XY)
        .build();

    assertArrayEquals(expected, actualB.row(index).yValues().toArray(), 0);
  }

}
